#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "syntax.h"
#include "box.h"


extern  double *xgrid, *ygrid, *zgrid;
extern  double *xcoord, *ycoord, *zcoord;

extern  int   xsize, ysize, zsize;
extern  long  xysize;

extern  int VERBOSE;
extern  int DIMENSIONALITY;

//***************************************************************
//                I M P L E M E N T A T I O N S
//***************************************************************

BoxObj::BoxObj(double x0, double x1, double y0, double y1, double z0, double z1)
{
  xmin = x0; xmax = x1;
  ymin = y0; ymax = y1;
  zmin = z0; zmax = z1; 
}

//***************************************************************

int BoxObj::inside(int ix, int iy, int iz)
  {    
  if (ix < 0 || ix >= xsize) return 0;
  bool b = (xcoord[ix] < xmax + 0.5 * xgrid[ix - ( ix > 0 ? 1 : 0)]) && 
           (xcoord[ix] > xmin - 0.5 * xgrid[ix]);

  if (DIMENSIONALITY > 1) {
    if (iy < 0 || iy >= ysize) return 0;
    b = b && (ycoord[iy] < ymax + 0.5 * ygrid[iy - ( iy > 0 ? 1 : 0) ]) && 
             (ycoord[iy] > ymin - 0.5 * ygrid[iy]);
  }
  if (DIMENSIONALITY > 2) {
     if (iz < 0 || iz >= zsize) return 0;
     b = b && (zcoord[iz] < zmax + 0.5 * zgrid[iz - ( iz > 0 ? 1 : 0) ]) && 
	      (zcoord[iz] > zmin - 0.5 * zgrid[iz]);
  }

  if (b) return _INSIDE_;  else return 0;
  }

//***************************************************************

int ObstBoxObj::inside(int ix, int iy, int iz)
  {    
  if (ix < 0 || ix >= xsize) return _INSIDE_;
  bool b = (xcoord[ix] < xmax - 0.5 * xgrid[ix]) && 
           (xcoord[ix] > xmin + 0.5 * xgrid[ix - ( ix > 0 ? 1 : 0)]);
  if (DIMENSIONALITY > 1) {
     if (iy < 0 || iy >= ysize) return _INSIDE_;
     b = b && (ycoord[iy] < ymax - 0.5 * ygrid[iy]) && 
	      (ycoord[iy] > ymin + 0.5 * ygrid[iy - ( iy > 0 ? 1 : 0) ]);
  }
  if (DIMENSIONALITY > 2) {
     if (iz < 0 || iz >= zsize) return _INSIDE_;
     b = b && (zcoord[iz] < zmax - 0.5 * zgrid[iz]) && 
      	      (zcoord[iz] > zmin + 0.5 * zgrid[iz - ( iz > 0 ? 1 : 0) ]);
  }

  if (b) return 0; else return _INSIDE_;
  }

//***************************************************************

int BoxObj::point_type(int ix, int iy, int iz)
    {
    int  surface;

    if ( (surface = inside(ix, iy, iz)) > 0 )
      {
      if (! inside(ix - 1, iy, iz) ) surface += SURF_XMIN;
      if (! inside(ix + 1, iy, iz) ) surface += SURF_XMAX;
      if (DIMENSIONALITY == 1) return surface;
      if (! inside(ix, iy - 1, iz) ) surface += SURF_YMIN;
      if (! inside(ix, iy + 1, iz) ) surface += SURF_YMAX;
      if (DIMENSIONALITY == 2) return surface;
      if (! inside(ix, iy, iz - 1) ) surface += SURF_ZMIN;
      if (! inside(ix, iy, iz + 1) ) surface += SURF_ZMAX;
      }

    return surface;
    }

//***************************************************************

int ObstBoxObj::point_type(int ix, int iy, int iz)
    {
    int surface;
    
    if ( (surface = inside(ix, iy, iz)) > 0 )
      {
      if (! inside(ix - 1, iy, iz) ) surface += SURF_XMIN;
      if (! inside(ix + 1, iy, iz) ) surface += SURF_XMAX;
      if (DIMENSIONALITY == 1) return surface;
      if (! inside(ix, iy - 1, iz) ) surface += SURF_YMIN;
      if (! inside(ix, iy + 1, iz) ) surface += SURF_YMAX;
      if (DIMENSIONALITY == 2) return surface;
      if (! inside(ix, iy, iz - 1) ) surface += SURF_ZMIN;
      if (! inside(ix, iy, iz + 1) ) surface += SURF_ZMAX;
      }

    return surface;
    }

//***************************************************************
//*   C L A S S   R E G I O N  I M P L E M E N T A T I O N S
//***************************************************************


RegionObj::RegionObj(int encls, int obstcls)
{
 encl_cnt = obst_cnt = 0;
 encl_num = encls;
 obst_num = obstcls;
 total = encl_num + obst_num;
 enclosures = new BoxObj[encl_num];
 obstacles = new ObstBoxObj[obst_num];
}

//***************************************************************

RegionObj::RegionObj(TokenString &pars)
{

 if (VERBOSE) { fprintf(stderr,"\n#### Building the geometry:"); fflush(stderr); }

 encl_cnt = obst_cnt = 0;
 encl_num = pars.token_count("volume");
 obst_num = pars.token_count("obstacle");

 total = encl_num + obst_num;

 if (!encl_num) 
   fprintf(stderr,"RegionObj::RegionObj: no geometry objects in parameter file!\n"), exit(1);

 enclosures = new BoxObj[encl_num];
 obstacles = new ObstBoxObj[obst_num];

 for (int i=1; i<=encl_num; i++)  add_enclosure(pars);
 if (obst_num)  for (int i=1; i<=obst_num; i++)  add_obstacle(pars);
}

//***************************************************************

RegionObj::~RegionObj()
{
  delete [] enclosures;
  delete [] obstacles;
}

//***************************************************************
// point_type bit fields are:
// bit # is:    ...8     |   7    |  6   |  5   |  4   |  3   |  2   |  1  
// meaning:   box number | inside | zmax | zmin | ymax | ymin | xmax | xmin
//***************************************************************


int RegionObj::point_type(int ix, int iy, int iz)
{
int    ptype, boxid = 0;
int    encl_bnd = SURF_MASK;
int    obst_bnd = SURF_MASK;  
int    inside = 0;

    for (int i = 0; i < encl_num; i++)
      {  
      ptype = enclosures[i].point_type(ix, iy, iz);
      if (ptype & _INSIDE_)    
        {                                      //  Point is inside if it is
	inside = _INSIDE_;                     // inside any of the enclosures.
        encl_bnd &= ptype;                     //  Inside point is on a boundary only
        if (ptype  & SURF_MASK )  boxid = i;   // if it belongs to the same boundary
	}                                      // of all the enclosures containing it
      }                                       
    if (encl_bnd == SURF_MASK)  encl_bnd = 0;  

    for (int j = 0; j < obst_num; j++)
      {  
      ptype = obstacles[j].point_type(ix, iy, iz);
      if (ptype == 0) inside = 0;              // Point is outside if it belongs
        else                                   // to any of the obstacles.
          if (ptype  & SURF_MASK )  
            { 
            obst_bnd &= ptype;                 // Inside point is on a boundary only
            boxid = j + encl_num;              // if it belongs to the same boundary
            if (obst_bnd == 0) inside = 0;     // of all the obstacles
            }                                  // Point is outside if it belongs to
      }                                        // boundaries of different orientations
    if (obst_bnd == SURF_MASK)  obst_bnd = 0;  

    if ( inside == 0 ) return 0;

    int bnd = encl_bnd + obst_bnd;
 
    if (bnd == 0) return inside;

    if ( ((bnd & SURF_XMIN) && (bnd & SURF_XMAX)) ||
         ((bnd & SURF_YMIN) && (bnd & SURF_YMAX)) ||
         ((bnd & SURF_ZMIN) && (bnd & SURF_ZMAX)) )
      
      { printf("RegionObj::point_type: ambiguity: surfaces touch\n"); exit(1); }

    else  return  _INSIDE_ + (boxid << 7) + bnd;
   
}


//***************************************************************

int RegionObj::point_type(long ind)
{
int iz = ind / xysize;
ind -= iz * xysize;
int iy = ind / xsize;
int ix = ind - iy * xsize;
return point_type(ix, iy, iz);
}

//***************************************************************

void RegionObj::add_enclosure(double x0, double x1, double y0, double y1,
                              double z0, double z1)
{
 if (encl_cnt >= encl_num) 
   {
   printf("*** RegionObj::add_enclosure: encl_cnt=%d >= encl_num=%d\n",
          encl_num, encl_cnt);
   return;
   }
 if (encl_cnt == 0) 
   { xmin = x0; xmax = x1; ymin = y0; ymax = y1; zmin = z0; zmax = z1; }
 else
   {
   if (x0 < xmin) xmin = x0;  if (x1 > xmax) xmax = x1;
   if (y0 < ymin) ymin = y0;  if (y1 > ymax) ymax = y1;
   if (z0 < zmin) zmin = z0;  if (z1 > zmax) zmax = z1;
   }

 if (VERBOSE) { fprintf(stderr,"   adding enclosure #%d:", encl_cnt+1); fflush(stderr); }

 enclosures[encl_cnt++] = BoxObj(x0, x1, y0, y1, z0, z1);

 if (VERBOSE) {
   fprintf(stderr," [%g,%g]", x0, x1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [%g,%g]", y0, y1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [%g,%g]", z0, z1);
   fprintf(stderr, "\n");
 }

 if ( x0 > x1 || y0 > y1 || z0 > z1 )
   fprintf(stderr,"*** ERROR: RegionObj::add_enclosure(): bad enclosure geometry [%g %g %g %g %g %g]\n%c",
           x0,x1,y0,y1,z0,z1, 7), exit(1);
}

//***************************************************************

void RegionObj::add_enclosure(TokenString &pars)
{
  double x0, x1, y0, y1, z0, z1;

  if (DIMENSIONALITY == 1)
    {
    pars.trail_pars("volume", encl_cnt+1, 'd', &x0, 'd', &x1, 'E');
    add_enclosure(x0, x1, 0.0, 0.0, 0.0, 0.0);
    }
  else  if (DIMENSIONALITY == 2)
    {
    pars.trail_pars("volume", encl_cnt+1, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');
    add_enclosure(x0, x1, y0, y1, 0.0, 0.0);
    }
  else
    {
    pars.trail_pars("volume", encl_cnt+1, 
     'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    
    add_enclosure(x0, x1, y0, y1, z0, z1);
    }
}

//***************************************************************

void RegionObj::add_obstacle(double x0, double x1, double y0, double y1, 
                             double z0, double z1)
{
 if (obst_cnt >= obst_num)  
   {
   printf("*** RegionObj::add_obstacle: obst_cnt=%d >= obst_num=%d\n",
          obst_num, obst_cnt);
   return;
   }

 if (VERBOSE) { fprintf(stderr,"   adding obstacle #%d:", obst_cnt+1); fflush(stderr); }

 obstacles[obst_cnt++] = ObstBoxObj(x0, x1, y0, y1, z0, z1);

 if (VERBOSE)  {
   fprintf(stderr," [%g,%g]", x0, x1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [%g,%g]", y0, y1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [%g,%g]", z0, z1);
   fprintf(stderr, "\n");
 }

 if ( x0 > x1 || y0 > y1 || z0 > z1 )
   fprintf(stderr,"*** ERROR: RegionObj::add_obstacle(): bad obstacle geometry [%g %g %g %g %g %g]\n%c",
           x0,x1,y0,y1,z0,z1,7), exit(1);
}

//***************************************************************

void RegionObj::add_obstacle(TokenString &pars)
{
  double x0, x1, y0, y1, z0, z1;

  if (DIMENSIONALITY == 1)
    {
    pars.trail_pars("obstacle", obst_cnt+1, 'd', &x0, 'd', &x1, 'E');
    add_obstacle(x0, x1, 0.0, 0.0, 0.0, 0.0);
    }
  else  if (DIMENSIONALITY == 2)
    {
    pars.trail_pars("obstacle", obst_cnt+1, 'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'E');
    add_obstacle(x0, x1, y0, y1, 0.0, 0.0);
    }
  else
    {
    pars.trail_pars("obstacle", obst_cnt+1, 
     'd', &x0, 'd', &x1, 'd', &y0, 'd', &y1, 'd', &z0, 'd', &z1, 'E');    
    add_obstacle(x0, x1, y0, y1, z0, z1);
    }
}

//***************************************************************
//                       T H E   E N D
//***************************************************************

void print_type(int x)
{
if (x == 0) printf("[OUT] ");
else
  {
  if (x & _INSIDE_) printf("[IN");
  if (!(x & SURF_MASK)) printf("] ");
    else
      {
      if (x & SURF_XMIN) printf(" XMIN");
      if (x & SURF_XMAX) printf(" XMAX");
      if (x & SURF_YMIN) printf(" YMIN");
      if (x & SURF_YMAX) printf(" YMAX");
      if (x & SURF_ZMIN) printf(" ZMIN");
      if (x & SURF_ZMAX) printf(" ZMAX");
      printf(" %d] ",x>>7);
      }
  }
}


