#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stream.h>
#include "Ygl.h"
#include "syntax.h"
#include "box.h"
#include "vector.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"
#include "fplot.h"
#include "pde.cc"
#include "global.cc"

//**************************************************************************************************

void header() {
     fprintf(stderr,"\n******************************************************");
     fprintf(stderr,"\n*                                                    *");
     fprintf(stderr,"\n*  Calcium diffusion/buffering solver: version 2.99  *");
     fprintf(stderr,"\n*    Victor Matveev, MRB/NIDDK/NIH, June 8, 2001     *");
     fprintf(stderr,"\n*                                                    *");
     fprintf(stderr,"\n******************************************************\n");
}

//**************************************************************************************************

#define  EXTRA_PARAM_STRING  "pA=5.182134 ; pi=4 atan(1) "

extern int VERBOSE;
extern int DIMENSIONALITY;
extern int GEOMETRY;
extern const int SPHERICAL;

//template <class T> void db(T mess) { cout << "debug message: " << mess << '\n'; }

//**************************************************************************************************

class LoopObj
{
public:

  PlotArray *plots;
  VectorObj *result;

  int       num, steps;
  VectorObj var, var0, dvar;
  char      **ids, **value, *tp, **res_ids;

  LoopObj(TokenString &, VectorObj *);
  ~LoopObj();
 
  int  operator()() { return num; }
  void step();
  void draw();
};

//**************************************************************************************************
//                                           M A I N
//**************************************************************************************************


main(int argc, char **argv) {

  if (argc < 2) { printf("*** Usage: %s parameter_file \n",argv[0]); exit(1); }
                     
  TokenString Params(argv[1], EXTRA_PARAM_STRING);    // Parse the parameter file   

  if (Params.token_count("verbose")) 
    if ( (VERBOSE = Params.trail_int("verbose")) > 6) { Params.verbose = 1; Params.print(); }

  if (VERBOSE) header();

  if (Params.token2_count("geometry","spherical")) { 
    DIMENSIONALITY = 1; GEOMETRY = SPHERICAL; } 
  else if  (Params.token2_count("geometry","conical")) { 
    DIMENSIONALITY = 2; GEOMETRY = 2; } 
 

  //************ 

  VectorObj result( Params.token_count("track") );
  LoopObj   vary( Params, &result );

  for (int i = 0; i <= vary.steps; i++)  // MAIN LOOP ++++++++++++++++++++++++++++
    {
    if ( !Params.token2_count("Run","ODE") )
      {
      RegionObj   synapse(Params);         // Read pre-synaptic geometry
      FieldObj::set_region(&synapse);      // Bind fields to geometry variable "synapse"
      set_bc_types(Params);                // Read boundary condition definitions
      set_dimensions(Params);              // read grid dimensions from parameter file
      grid_stretch(Params);                // make non-uniform grid (read parameters from file)
      grid_clamp(Params);                  // clamp grid nodes to specified locations
      initScheme();                        // set up finite difference support arrays
      FieldObj Ca(Params);                 // set up calcium field (read parameters from file)
      BufferArray Bufs( Params);           // set up buffer fields (read parameters from file)
      KineticObj Gates(Params, &Ca, &Bufs);// set up Ca binding kinetic variables
      if ( Params.token_count("Import") ) Import( Params.trail_string("Import"), &Gates, &Ca, &Bufs);

      PlotArray *plots = vary() ? new PlotArray(0) : new PlotArray(Params, &Ca, &Bufs, &Gates);

      if (!Ca.source_num) fprintf(stderr,"\n>>> Error: no Ca channels specified <<<\n"), exit(1);
      Run(Params, Ca, Bufs, Gates, *plots);//##### RUN THE DIFFERENCE SCHEME #####
      if ( vary() ) for (int i = 0; i < result.size; i++) 
        result[i] = KinExpression(Params, Params.token_index("track", i + 1) + 1, 0, 
                                  &Ca, &Bufs, &Gates).Evaluate();
      delete plots;
      kill_grid();
      kill_bc();
      }
    else   //#####################  ODE solver mode: #################
      {
      KineticObj Gates(Params, 0, 0);            // set up ODEs
      if ( Params.token_count("Import") ) Import( Params.trail_string("Import"), &Gates, 0, 0);
      PlotArray *plots = vary() ? new PlotArray(0) : new PlotArray(Params, 0, 0, &Gates);
      RunODEs(Params, Gates, plots);
      if ( vary() ) for (int i = 0; i < result.size; i++) 
        result[i] = KinExpression(Params, Params.token_index("track", i + 1) + 1, 0,
                                  0, 0, &Gates).Evaluate();
      delete plots;
      }

    vary.draw(); 
    if (i < vary.steps) vary.step();
    }                                   

}

//**********************************************************************************************
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//**********************************************************************************************

LoopObj::LoopObj(TokenString &TS, VectorObj *res) : num( TS.token_count("for") ), 
var(num), var0(num), dvar(num), ids(new char *[num]), value(new char *[num]), tp(new char[num])
  {
  int i;
  if (!num) { steps = 0; return; }

  if (!res->size) 
     fprintf(stderr,"## LoopObj::LoopObj: no variables to track in the \"for\" loop\n"), exit(1);

  fprintf(stderr, "\n #### Running the simulation for multiple values of parameter(s):\n");

  int BUFFER = 100;
  steps = 100000;
  char prefix[256], filename[256];

  strcpy(prefix, "");
  result = res;

  char   id[80], seq[3], sto[3], sstep[5]; 
  double v0, v1, dv = 1.0;

  res_ids = new char *[res->size + 1];

  for (i = 0; i < res->size; i++)
      {       
      char s[1025]; strcpy(s,"");
      int p = TS.token_index("track",i+1);
      while (!TS.equal(++p, TS.carr_ret)) strcat(s,TS[p]);
      res_ids[i] = StrCpy(s);
      }

  for (i = 0; i < num; i++)
      {       
      strcpy(sstep,"step");
      TS.trail_pars("for", i+1, 0, 0, 's', seq, 'd', &v0, 's', sto, 'd', &v1, 's', sstep, 'd', &dv);
      strcpy(id, TS[ TS.token_index("for", i+1) + 1 ]);      

      if (!equal(seq,"=") || !equal(sto,"to") || !equal(sstep,"step") ) { 
        printf("LoopObj::LoopObj: syntax error: [for %s %s %g %s %g %s %g]\n", 
               id, seq, v0, sto, v1, sstep, dv); 
        printf(" Usage: for x = x0 to x1 step dx\n");
        exit(1); }

      if ( ( fabs(v0 - int(v0+0.5)) < 1.0e-6 ) && 
           ( fabs(v1 - int(v1+0.5)) < 1.0e-6 ) &&  
           ( fabs(dv - int(dv+0.5)) < 1.0e-6 ) )  
      tp[i] = 'i'; else tp[i] = 'd';

      ids[i] = StrCpy(id);

      var0.elem[i] = v0;
      dvar.elem[i] = dv;
   
      int n = int( (v1 - v0) / dv + 1e-10 );
      if (n < steps) steps = n;
    
      int ix = TS.token2_index(id, "=") + 1;
      value[i] = new char[BUFFER];

      if (tp[i] == 'i')
        sprintf(value[i], "%d", int(v0 + 0.5) );
      else
        sprintf(value[i], "%g", v0 );

      TS.token_ptr[ix] = value[i];  // hijack the token pointer
    
      strcpy(TS.token_ptr[ix+1],TS.carr_ret);  // replace "to" with carriage return

      fprintf(stderr, " ## For loop: for(%s = %g; %s <= %g; %s += %g)\n",
	      ids[i], v0, ids[i], v1, ids[i], dv);
      }
  var = var0;

  if ( TS.token2_count("plot_method","xmgr") ) {
     plots = new PlotArray(2 * result->size);
     plots->method = METHOD_XMGR;
     XmgrPlot::init(result->size);
  }
  else if (TS.token2_count("plot_method","mute")) {
     plots = new PlotArray(result->size);
     plots->method = METHOD_MUTE;
  }
  else { 
     plots = new PlotArray(2 * result->size);
     plots->method = METHOD_YGL; 
     plots->ygl_on = 1;
     strcat(id,"=");
  }
 
  if (TS.token_count("plot_print")) TS.trail_pars("plot_print", 1, 's', prefix, 'E');

  for (i = 0; i < result->size; i++) {
    fprintf(stderr, " ### tracking variable #%d: %s\n", i+1, res_ids[i]);
    sprintf(filename,"%s%s", prefix, res_ids[i]);

    if (plots->method == METHOD_YGL) {
      plots->set_plot( new PointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps, res_ids[i], id, 
                                       pref_width, pref_height, res_ids[i]),"point");
      plots->set_plot( new MutePointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps, filename, res_ids[i]), "point file");
      }
    else if ( plots->method == METHOD_MUTE )
      plots->set_plot( new MutePointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps,  filename, res_ids[i]), "point file");
    else if ( plots->method == METHOD_XMGR ) {
      plots->set_plot( new XmgrPointPlot(1, 0, steps * dv, id), "point xmgr");
      ((XmgrPointPlot *)(plots->array[2*i]))->set_set( (*result)()+i, var()+num-1, res_ids[i]);
      plots->set_plot( new MutePointPlot( (*result)()+i, var()+num-1, 0, steps * dv, steps,  filename, res_ids[i]), "point file");
    }
  }
  }

//**************************************************************************************************
//**************************************************************************************************

  LoopObj::~LoopObj()
  {
   int i;
   if (!num) return;

   delete plots;
   delete [] tp;

   for (i=0; i<result->size; i++) delete res_ids[i];
   for (i=0; i<num; i++) { delete ids[i]; delete value[i]; }

   delete [] value; delete [] ids; delete [] res_ids;
 
   var -= dvar;

  }


//**************************************************************************************************

  void LoopObj::step()
  {
  int i;
  if (!num) return;

  var += dvar;
  for (i = 0; i < num; i++)
    {       
    if (tp[i] == 'i')
      sprintf(value[i], "%d", int(var[i] + 0.5) );
    else
      sprintf(value[i], "%g", var[i] );
    }
 
    fprintf(stderr,"\n=================================================\n");
    fprintf(stderr,"||   Next step in the \"for\" loop: ");
    for (i = 0; i < num; i++)  fprintf(stderr,"%s=%s ", ids[i], value[i]);
    fprintf(stderr,"\n=================================================\n");

   }

//**************************************************************************************************

  void LoopObj::draw()
    {
    int i;
    if (!num) return;

    plots->draw_all();  

    fprintf(stderr,"=================================================\n ");
    for (i = 0; i < num; i++)  fprintf(stderr,"%s=%s ", ids[i], value[i]);
    fprintf(stderr,":");
    for (i = 0; i < result->size; i++) fprintf(stderr," %s=%g", res_ids[i], (*result)[i]);
    fprintf(stderr,"\n=================================================\n");

    fflush(stdout);
    }

//**************************************************************************************************
