#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"

#define NUMBER_TYPE   30
#define POINTER_TYPE  31
#define VARIABLE_TYPE 32


#define T_PLUS  0
#define T_MINUS 1
#define T_MULT  2
#define T_DIV   3

#define BRACKET 4
#define BR_CLOSE 5

#define SINH 6
#define COSH 7
#define TANH 8
#define SIN  9
#define COS  10
#define TAN  11
#define EXP  12
#define SQR  13
#define SQRT 14
#define THETA 15
#define STEP  16
#define LOG   17
#define LOG10 18
#define CUBE  19
#define ATAN  20
#define TYPE_NUM 21

extern int  VERBOSE;
extern long location_to_index(double, double, double);

const char type_id[TYPE_NUM][7] = { "+", "-", "*", "/", "(", ")", 
                        "sinh", "cosh", "tanh", "sin", "cos", "tan", 
                        "exp", "sqr", "sqrt", "theta", "step", "log", "log10", "cube", "atan" };

//**************************************************************************

void ExpressionObj::Optimize()
{
  if (term_num == 0) return;

  for (int term=0; term < term_num; term++)
      if ( term_array[term].type == NUMBER_TYPE)
        {
        int tt = term;
        double v = term_array[tt].val;
        while ( (term_array[tt+1].type == NUMBER_TYPE) && (tt + 1 < term_num) )
          { v *= term_array[tt+1].val; tt++; }
        if (tt > term)
  	  {
          int dt = tt - term;
          term_array[term].val = v;
          term_num -= dt;
          for (tt = term + 1; tt < term_num; tt++)        
	    term_array[tt] = term_array[tt + dt];
	  }
	}
}

//**************************************************************************

inline double plus_minus(char op, double &a, double b)
{
  if  (op == T_PLUS) return  a += b;
    else if (op == T_MINUS) return a -= b;
  else printf("!!! Operation [%d=%c] not recognized\n", op, op);
  exit(1);
}

//**************************************************************************

inline double mult_div(char op, double &a, double b)
{
  if  (op == T_MULT) return  a *= b;
    else if (op == T_DIV) return a /= b;
  else printf("!!! Operation [%d=%c] not recognized\n", op, op);
  exit(1);
}

//**************************************************************************

double function(char op, double x)
{
  switch (op)
    {
    case BRACKET: return x;
    case COSH:    return cosh(x);
    case SINH:    return sinh(x);
    case COS:     return cos(x);
    case SIN:     return sin(x);
    case TANH:    return tanh(x);
    case TAN:     return tan(x);
    case ATAN:    return atan(x);
    case EXP:     return exp(x);
    case LOG:     return log(x);
    case LOG10:   return log10(x);
    case SQR:     return x * x;
    case CUBE:    return x * x * x;
    case SQRT:    return sqrt(x);
    case THETA:   if (x == 0.0 ) return 0.5; else return (x > 0) ? 1 : 0;
    case STEP:    return 0.5 * (1 + tanh(x) );
    default:      printf("function: Function [%c] not recognized\n", op);
                  exit(1);
    }
}

//**************************************************************************

double ExpressionObj::Evaluate()
  {
  const    int LEVELS = 5;
  double   result[LEVELS], value[LEVELS];
  unsigned char  plusmin[LEVELS], multdiv[LEVELS], func[LEVELS], first[LEVELS];
  char     level;

  struct TermStruct term;

  level = 0;
  result[0] = 0.0;    value[0] = 0.0;
  plusmin[0] = T_PLUS;  multdiv[0] = T_MULT;
  func[0] = BRACKET;  first[0] = 1;

  for (int j=0; j < term_num; j++)

    switch ( (term = term_array[j]).type )
      {
      case T_PLUS:        plus_minus(plusmin[level], result[level], value[level]);
	                  value[level] = 1; plusmin[level] = T_PLUS; 
                          break;

      case T_MINUS:       plus_minus(plusmin[level], result[level], value[level]);
	                  value[level] = 1; plusmin[level] = T_MINUS; 
                          break;

      case T_DIV:         multdiv[level] = T_DIV; break;

      case T_MULT:        break;

      case BR_CLOSE:      result[level] = 
                          function(func[level], plus_minus(plusmin[level], result[level], value[level]) );
	                  level --;
                          if (first[level]) { value[level] = result[level + 1]; first[level] = 0; }
                            else mult_div(multdiv[level], value[level], result[level + 1] );
                          multdiv[level] = T_MULT;
                          break;

      case NUMBER_TYPE:   if (first[level]) { value[level] = term.val; first[level] = 0; }
                            else mult_div(multdiv[level], value[level], term.val );
                          multdiv[level] = T_MULT;
                          break;

      case POINTER_TYPE:
      case VARIABLE_TYPE: if (first[level]) { value[level] = *term.ptr; first[level] = 0; }
                            else mult_div(multdiv[level], value[level], *term.ptr );
                          multdiv[level] = T_MULT;
                          break;

      default:          // opening bracket or function
         	if (term.type >= TYPE_NUM) 
		  {
		  fprintf(stderr,"ExpressionObj::Evaluate: function [%d] not recognized (term %d of %d)\n",
                          term.type,j,term_num);
                  exit(1);
		  }
                if (term.type != BRACKET) j++;
	        if ( (++level) > LEVELS ) 
                  fprintf(stderr, "ExpressionObj::Evaluate: exceeded bracket nesting limit\n"), exit(1);
                result[level] = 0.0;      value[level] = 0.0;
                plusmin[level] = T_PLUS;    multdiv[level] = T_MULT;
                func[level] = term.type;  first[level] = 1;  break;
      }

  if (level != 0) { fprintf(stderr, "ExpressionObj::Evaluate: nesting error: level=%d\n",level); exit(1); }

  return  function(func[0], plus_minus(plusmin[0], result[0], value[0]) );
  }

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//**************************************************************************

KinExpression::KinExpression(TokenString &Param, int p, double *t, 
                             FieldObj *Ca, BufferArray *Bufs, KineticObj *KO) :
  ExpressionObj( get_term_num(Param, p) )
{
  int tp;

  for (int term = 0; term < term_num; term++, p++)
    {
    for (tp = 0; tp < TYPE_NUM; tp++)
      if (Param.equal(p, type_id[tp])) 
        { 
        term_array[term].type = tp; 
        break; 
        }

    if (tp < TYPE_NUM) continue;

    if (Param.equal(p,"t") && t)
      {
      term_array[term].type = POINTER_TYPE;
      term_array[term].ptr = t;
      term_array[term].val = 0;
      continue;
      }

    if (Param.is_number(p)) 
      { 
      term_array[term].type = NUMBER_TYPE;
      term_array[term].ptr = 0;
      term_array[term].val = Param.index_double(p);
      continue;
      }

    term_array[term].type = POINTER_TYPE;

    if ( !( term_array[term].ptr = ResolveID(Param, p, 0, Ca, Bufs, KO) ) )
      fprintf(stderr, "KinExpression::KinExpression: cannot interpret [%s]\n",Param[p]), exit(1);

    term_array[term].val = 0;
    }

  if (VERBOSE > 6) 
    { fprintf(stderr,"\nExpression interpreted: "); print(stderr, Ca, Bufs, KO); }
  // Optimize(); 
  // print(stderr,Ca,Bufs,KO); fprintf(stderr,"\n");
}

//**************************************************************************


void ExpressionObj::print(FILE *f, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
  char *s; 

  for (int term=0; term < term_num; term++)
   switch( term_array[term].type )
     {
     case NUMBER_TYPE:  fprintf(f,"%g ",term_array[term].val);
                        break;
     case POINTER_TYPE: if (s = ResolvePtr(term_array[term].ptr, Ca, Bufs, KO) )
                            fprintf(f,"%s ", s);
                        else fprintf(f,"t ");   
                          // correct above in the future: any way to tell if this pointer is really time?
			  // fprintf(f,"[%X] ", term_array[term].ptr );
                        break;
     default:           fprintf(f,"%s ", type_id[term_array[term].type]);
                        break;
     }
 fflush(f);
}

//**************************************************************************

int get_term_num(TokenString &Param, int p)
{
  int terms = 0;
  while (! (Param.equal(p, Param.carr_ret) || Param.equal(p, "to")) ) { p++; terms++; }
  return terms;
}

//**************************************************************************

double *ResolveID(TokenString &Param, int p, double **t, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
   int j;

  if (VERBOSE > 8) fprintf(stderr, ">> ResolveID: Resolving %s\n",Param[p]);
  char svar[80];
  Param.index_string(p, svar);

   if (Ca)
     if ( equal(svar, Ca->ID) )  
       { if (t) *t = &Ca->Time; return (*Ca)(); }

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
	if ( equal(svar,Bufs->array[j]->ID) )  
          { if (t) *t = &Bufs->array[0]->Time;  return (*Bufs->array[j])(); }

    if (KO)
      {
      if (t) *t = &KO->Time;

      if (KO->var) 
        for (j = 0; j < KO->var->get_size(); j++)
          if (equal(svar, KO->var_id[j])) 
            return (*KO->var)() + j;

      if (KO->ident)
        for (j = 0; j < KO->ident->get_size(); j++)
          if (equal(svar, KO->ident_id[j])) 
            return  (*KO->ident)() + j;

      for (j = 0; j < KO->tables->table_num; j++)
	 if (equal(svar, KO->tables->array[j]->ID ) )
           return &(KO->tables->array[j]->value);

      for (j = 0; j < KO->facilitation->peak_num; j++)
	 if (equal(svar, KO->facilitation->array[j]->ID ) ) 
           return &(KO->facilitation->array[j]->peak);

      for (j = 0; j < KO->locations->interpol_num; j++)
	 if (equal(svar, KO->locations->array[j]->ID ) ) 
	   {
           if (t) *t = &(KO->locations->array[j]->eigentime);
	   return &(KO->locations->array[j]->result);
	   }
      }

    fprintf(stderr,"ResolveID: cannot interpret [%s]\n", svar ), exit(1);
    return 0;
}

//**************************************************************************


char *ResolvePtr(double *ptr, FieldObj *Ca, BufferArray *Bufs, KineticObj *KO)
{
int j;
    if (Ca)  if ( ptr == Ca->elem )  return Ca->ID; 

    if (Bufs)
      for (j = 0; j < Bufs->buf_num; j++)
	if ( ptr == Bufs->array[j]->elem )  return  Bufs->array[j]->ID; 

    if (KO)
      {
      if (KO->var) 
        for (j = 0; j < KO->var->get_size(); j++)
         if ( ptr == ((*KO->var)() + j) ) return KO->var_id[j];
          
      if (KO->ident)
        for (j = 0; j < KO->ident->get_size(); j++)
         if ( ptr == ((*KO->ident)() + j) ) return KO->ident_id[j];

      for (j = 0; j < KO->tables->table_num; j++)
	 if ( ptr == &(KO->tables->array[j]->value) ) return KO->tables->array[j]->ID;

      for (j = 0; j < KO->facilitation->peak_num; j++)
	 if ( ptr == &(KO->facilitation->array[j]->peak) ) return KO->facilitation->array[j]->ID;

      for (j = 0; j < KO->locations->interpol_num; j++)
	 if ( ptr == &(KO->locations->array[j]->result) ) return KO->locations->array[j]->ID;
      }

    return 0;
    //fprintf(stderr,"ResolvePtr: cannot resolve pointer %X\n", ptr );
    //exit(1);
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


