#ifndef FPLOT_H
#define FPLOT_H

#define METHOD_YGL  0
#define METHOD_XMGR 1
#define METHOD_MUTE 2

#define UPDATE_FRAC 0.001  // fraction of the min-max amplitude change required to update mute plot

#define UPDATE_STEPS 600
#define XMGR_STEPS 600

#define    xmarg_color_panel 50
#define    ymarg_color_panel 15
#define    dy_color_panel 10
#define    color_num 150

#define pref_width 380
#define pref_height 200
#define xmargin 15
#define ymargin 15
#define char_height 8
#define char_width  6

#define bgr_color 0
#define plot_color 1
#define grid_color 2
#define text_color 3
#define label_color 4

#define pixelspergridline 45
#define line_color 5

#define dxmargin 0.14
#define dymargin 0.21

#define dx_legend 0.2
#define dy_legend 0.1

#define title_color 4  // landscape mode

void   mouse_wait();
void   mouse_pause();

double format_double(double number, int divisions, int sigdigits, char *fstr);
void   format_print(double number, int sigdigits, char *strout);

void Export(const char *filename, KineticObj *KO, FieldObj *Ca, BufferArray *BA);
void Import(const char *filename, KineticObj *KO, FieldObj *Ca, BufferArray *BA);

extern double get_sim_time(TokenString &params, int &sim_steps);
extern int    VERBOSE;

//*******************************************************************************
//                 C L A S S   P L O T O B J E C T
//*******************************************************************************

class PlotObj
{
protected:
public:

  int   plot_height, win_height, plot_width,  win_width;
  char  win_title[255];
  long  winid;

  char     x_label[80];
  double   x_value;

  char  log_plot, mute;

  double   fmin, fmax;
  double   *fptr;

  PlotObj(double *ptr, char islog, const char *txt = 0, const char *label = 0)
    {
    fptr = ptr; log_plot = islog;
    if (txt) strcpy(win_title,txt); else win_title[0] = 0;
    if (label) strcpy(x_label,label); else x_label[0] = 0;
    x_value = fmax = fmin = 0.0;
    };

  virtual void draw() = 0;
  virtual void redraw() = 0;
  virtual void print() = 0;
  void    draw_grid(double, double, double, double, char);

  double get_value(long ind = 0)
   {
   double v = fptr[ind];
   if (!log_plot) return v;
      else return ( (v <= 0.0) ? 0.0 : log(v)/log(10.0) );
   };

};

//*******************************************************************************
//                 C L A S S   F I E L D P L O T O B J E C T
//*******************************************************************************

class FieldPlotObj : public PlotObj
{
protected:
public:

 FieldObj *field; 
 long     field_index;

 FieldPlotObj(FieldObj *f, char islog, const char *txt = 0, const char *label = 0):
   PlotObj((*f)(), islog, txt, label)  { field = f; };


};


//*******************************************************************************
//                      C L A S S   2 D - P L O T
//*******************************************************************************

class Plot2D : public FieldPlotObj
{
protected:

        int color1;  

  long     incr1, incr2, num1, num2;

  char    fixed;
  double  max_fixed, min_fixed;
  double  df_dcol, df_dcol_fixed;

  double  *grid1, *grid2;
  double  *coord1, *coord2;

  double tscale;

public:

 Plot2D(FieldObj *f, char islog, char dir, double coord, double, int, 
        int hor = pref_width, int ver = pref_height, double min = 0.0, double max = -1.0,
        const char *label = "time=");

 void    draw();
 void    redraw();
 void    set_colors();
 void    get_range();
 int     get_color(double);
 void    color_panel();
 void    print() { return; }

};

//*******************************************************************************
//                      C L A S S   1 D - P L O T
//*******************************************************************************

class FieldPlot1D : public FieldPlotObj
{
protected:

  long    incr, num;
  double  *grid, *coord;

public:

 FieldPlot1D(FieldObj *f, char islog, char dir, double coord1, double coord2, 
             const char *label = "time=");

 void    get_range();
};

//*******************************************************************************

class Plot1D : public FieldPlot1D
{
protected:

  double tscale;

public:

 Plot1D(FieldObj *f, char islog, char dir, double coord1, double coord2,
        double T, int n = UPDATE_STEPS, int hor = pref_width, int ver = pref_height, 
        const char *label = "time=");

 void    draw();
 void    redraw() { x_value = 0.0; draw(); }; 
 void    set_colors();
 void    get_range();
 void    print();

};

//*******************************************************************************
//                  C L A S S   R A S T E R   P L O T
//*******************************************************************************

class RasterPlot : public FieldPlot1D
{
protected:

  double *raster;
  double *times;
  double last;
  int    steps, count;

  int color1;  

  char    fixed;
  double  max_fixed, min_fixed;
  double  df_dcol, df_dcol_fixed;

public:

 RasterPlot(FieldObj *f, char islog, char dir, double coord1, double coord2, double t, int n,
  int hor = pref_width, int ver = pref_height, double min = 0.0, double max = -1.0,
  const char *label = "time=");

 ~RasterPlot() {  if (VERBOSE > 7) { fprintf(stderr,"~RasterPlot() "); fflush(stderr); }
                  delete raster; delete times; };

 void    draw();
 void    redraw();
 void    set_colors();
 void    get_range();
 int     get_color(double);
 void    color_panel();
 void    print() { return; }

};

//*******************************************************************************
//                  C L A S S   P O I N T   P L O T
//*******************************************************************************

class PointPlot : public PlotObj
{
protected:

  double *buffer;
  double *times;
  double *timeptr;
  double last;
  int    steps, count;

  const char *fname;

public:

 PointPlot(double *ptr, double *tptr, char islog, double t, int n,
          const char *txt = 0, const char *label = "time=", 
          int hor = pref_width, int ver = pref_height, const char *filename = "" );

 ~PointPlot() { if (VERBOSE > 7) { fprintf(stderr,"~PointPlot() "); fflush(stderr); }
                delete [] buffer; delete [] times; };

 void    draw();
 void    print();
 void    redraw()  { if (count) count --; draw(); }

 void    set_colors();
 void    get_range();

};

//*******************************************************************************
//               C L A S S   M U T E   P O I N T   P L O T
//*******************************************************************************

class MutePointPlot : public PlotObj
{
protected:

  double *time;
  double fold;
  double tscale;
  char   *filename;

public:

 FILE   *file;

 MutePointPlot(double *ptr, double *tptr, char islog,  double t, int n,
               const char *fname = 0, const char *txt = 0) :
 PlotObj(ptr, islog, txt)
    {
    tscale = double(n) / t;
    fmin = fold = fmax = get_value();
    x_value = -1e12;
    time = tptr;
    if ( equal(fname, "stdout") || equal(fname, "") ) file = stdout;
    else if ( !(file = fopen(filename = StrCpy(fname), "w")) ) 
      { printf("MutePointPlot: cannot open [%s] \n", fname); exit(1); }
    };

 ~MutePointPlot() { if (VERBOSE > 7) { fprintf(stderr,"~MutePointPlot() "); fflush(stderr); }
                    fclose(file); delete [] filename; };

 void    draw();
 void    redraw()  { x_value = 0; draw(); }
 void    print() { return; }
};

//*******************************************************************************

class MutePlot1D : public FieldPlot1D
{
public:

 char   *file_name;
 double total;
 double tscale;

 MutePlot1D(FieldObj *f, char islog, char dir, double coord1, double coord2, double T,
            const char *fname = 0, const char *txt = 0) :
 FieldPlot1D(f, islog, dir, coord1, coord2)
   {
   tscale = double(UPDATE_STEPS) / (total = T);
   file_name = StrCpy(fname);
   };

 ~MutePlot1D() { if (VERBOSE > 7) { fprintf(stderr,"~MutePlot1D() "); fflush(stderr); }
                 delete file_name; };

 void    draw() {  FILE *file;
                   double t = field->Time;
                   if ( int(t * tscale) == int(x_value * tscale) ) return;
                   x_value = t;
                   long ind = field_index;
                   if ( !(file = fopen(file_name, "w")) ) 
                      printf("MutePlot1D: cannot open [%s] \n", file_name);
                   for (int j = 0; j < num; j++, ind += incr) 
                      fprintf(file,"%.10g %.10g \n",coord[j],get_value(ind));
                   fclose(file);
                }

 void    redraw() { x_value = 0.0; draw(); }; 
 void    print() { return; }

};

//*******************************************************************************

class FieldDump : public FieldPlotObj
{
protected:

  double dt, t0, t1;
  char   filename[255];

public:

 FILE   *file;

 FieldDump(FieldObj *f, double t0_, double t1_, double dt_, const char *fname = 0, 
           const char *txt = 0) : FieldPlotObj(f, 0, txt)
    {
    strcpy(filename, fname); strcat(filename, field->ID);
    t0 = t0_; t1 = t1_; dt = dt_;
    x_value = t0;
    };

 // ~FieldDump() { }

 void    draw()
   {
   char fname[255];

   if ( (field->Time < x_value - 1e-12) || (x_value > t1 + 1e-12) ) return;
   sprintf(fname,"%s.%.4g",filename, x_value);
   field->Export(fname);
   x_value += dt; 
   };

 void    redraw() { draw(); }
 void    print()  { return; }
};

//*******************************************************************************

class StateDump : public PlotObj
{
protected:

  double dt, t0, t1;
  char   filename[255];
  FieldObj    *CA;
  BufferArray *BA;
  KineticObj  *KO;

public:

 FILE   *file;

 StateDump( KineticObj *ko, FieldObj *f, BufferArray *ba, 
           double t0_, double t1_, double dt_, const char *fname = 0, 
           const char *txt = 0) : PlotObj(0, 0, txt)
    {
    strcpy(filename, fname); 
    CA = f; BA = ba; KO = ko;
    t0 = t0_; t1 = t1_; dt = dt_;
    x_value = t0;
    };

 // ~FieldDump() { }

 void    draw()
   {
   char fname[255];

   if ( (KO->Time < x_value - 1e-12) || (x_value > t1 + 1e-12) ) return;
   sprintf(fname,"%s.%.4g",filename, x_value);
   Export(fname, KO, CA, BA);
   x_value += dt; 
   };

 void    redraw() { draw(); }
 void    print()  { return; }
};

//*******************************************************************************
//           B A S E   C L A S S   X M G R   P I P E   P L O T
//*******************************************************************************
// xmgr colors: 0-white 1-black 2-red 3-green 4-blue 5-yellow 6-brown 7-gray 
// 8-violet 9-cyan 10-magenta 11-orange 12-indigo 13-maroon 14-turquise 15 -green4

class XmgrPlot
{
protected:


  static int graph_num;
  static int graph_count;

  static int rows;
  static int cols;

  static FILE *file;

  int    set_num;
  int    set_count;
  int    graph_id;

  int    view_left, view_right, view_top, view_bottom;

  int    steps;
  double first, last;

public:

 static void init(int gnum, FILE *f = stdout, int = 0, int = 0);

 XmgrPlot(int sets, char is_log, const char *xlabel = "time");

 ~XmgrPlot() { if (VERBOSE > 7) fprintf(stderr,"~XmgrPlot() "), fflush(stderr);  };

};

//*******************************************************************************
//               C L A S S   X M G R   P I P E   P L O T
//*******************************************************************************
// xmgr colors: 0-white 1-black 2-red 3-green 4-blue 5-yellow 6-brown 7-gray 
// 8-violet 9-cyan 10-magenta 11-orange 12-indigo 13-maroon 14-turquise 15 -green4

class XmgrPointPlot : public XmgrPlot, public PlotObj
{
protected:


  double **tptrs, **fptrs;
  double *x_old;
  double *f_old;

public:

 XmgrPointPlot(int sets, char islog, double T = 0, const char *xlabel = "time");

 void set_set(double *p, double *t, const char *txt = 0);

 ~XmgrPointPlot() { if (VERBOSE > 7) { fprintf(stderr,"~XmgrPointPlot() "); fflush(stderr); }
                    delete [] tptrs; delete [] fptrs; delete [] x_old; delete [] f_old; 
                   } 

 void    get_range();
 void    draw();

 void    redraw()  { for (int i = 0; i < set_num; i++) x_old[i] = 0.0; draw(); }

 void    print() { return; }
};


//*******************************************************************************
//            C L A S S   1 D   X M G R   P I P E   P L O T  
//*******************************************************************************

class Xmgr1Dplot : public XmgrPlot, public FieldPlot1D
{
protected:

public:

 Xmgr1Dplot(FieldObj *f, char islog, char dir, double coord1, double coord2, double T, 
            int n = 100, int m = 5);

 void    draw();
 void    redraw() { x_value = 0.0; draw(); }; 
 int     set_color(int n) ;
 void    set_colors();
 void    get_range();
 void    print() { return; }

};

//*******************************************************************************
//                      C L A S S   P L O T   A R R A Y
//*******************************************************************************

class PlotArray
{
 protected:

 public:

  PlotObj **array;
  char    **plotype;
  int     plot_num;
  int     count;
  char    ygl_on, method;

  PlotArray(TokenString &, FieldObj *, BufferArray *, KineticObj *);

  int get_plot_num(TokenString &);

  PlotArray(int n)
     { 
     count = 0; ygl_on = 0;
     plot_num = n;
     if (n) { array = new PlotObj *[n]; 
              plotype  = new char *[n]; }
     }

  ~PlotArray();

  void set_plot(PlotObj *plot, const char *type_id) 
     {
     if (count >= plot_num) 
       printf("*** Error: PlotArray: set_plot: count: %d >= %d\n", count, plot_num); 
     else  
       {
       array[count] = plot;
       plotype[count] = StrCpy(type_id);
       count ++;
       }
     }

  void draw_plot(int n)   { if (n < count) array[n]->draw();  }

  void redraw_plot(int n)  { if (n < count) array[n]->redraw(); }

  void draw_all() 
    {  if (count) for (int i=0; i<count; i++) array[i]->draw(); }

  void redraw_all()
    {  if (count) for (int i=0; i<count; i++) array[i]->redraw();  }

  void print_all()
    {  if (count) for (int i=0; i<count; i++) array[i]->print(); }

  void tail_draw() 
    { 
    while(ygl_on)
      {
      while(! (getbutton(RIGHTMOUSE) | getbutton(LEFTMOUSE) | getbutton(MIDDLEMOUSE) ) );
      if ( getbutton(RIGHTMOUSE) ) return;
      redraw_all(); 
      while( getbutton(RIGHTMOUSE) | getbutton(LEFTMOUSE) | getbutton(MIDDLEMOUSE) );
      }
    }
};

//*******************************************************************************

#endif
