#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"
#include "Ygl.h"
#include "fplot.h"

extern int VERBOSE;

//**************************************************************************

void KineticObj::initialize()
    {
    var    = new VectorObj(var_num);
    varOld = new VectorObj(var_num);
    var_id = new char *[var_num];

    ident = new VectorObj(ident_num);
    ident_id = new char *[ident_num];

    formulas = new KinExpression *[var_num + ident_num];
    }

//**************************************************************************

  KineticObj::~KineticObj()  
    {
    int i;
    if (VERBOSE > 7) { fprintf(stderr,"> Destructing ODEs and variables: ");
                       fflush(stderr); }

    delete facilitation;

    for (i=0; i < ident_num; i++)
      {
      delete ident_id[i];
      delete formulas[i];
      }

    for (i=0; i < var_num; i++)
      {
      delete var_id[i];
      delete formulas[i+ident_num];
      }

    delete var;
    delete varOld;
    delete [] var_id;

    delete ident;
    delete [] ident_id;

    delete [] formulas;

    delete locations;
    delete tables;
    if (VERBOSE > 7) fprintf(stderr," done\n");
 }

//**************************************************************************


KineticObj::KineticObj(TokenString &Param, FieldObj *Ca, BufferArray *Bufs)
{
   int i;

  int var_num1 = Param.token_count(ODE_LABEL);
  int var_num2 = Param.token3_count("/","dt","=");
  var_num      = var_num1 + var_num2;

  ident_num = Param.token_count(IDENT_LABEL);
  initialize();  

  Time = 0.0;
  locations    = new InterpolArray(Param, Ca, Bufs);
  tables       = new TableArray(Param);
  facilitation = new PeakTrackArray(Param);

  for (i = 0; i < var_num; i++) {
    if (i < var_num1) var_id[i] = Param.StrCpy( Param.token_index(ODE_LABEL, i+1) - 1 );
      else var_id[i] = StrCpy( Param[ Param.token3_index("/","dt","=", i+1-var_num1) - 3 ] + 1 );
    (*var)[i] = 0.0;
    for (int j = 0; j < Param.token3_count("0",")","="); j++) {           // bootstrap!!!
      int p = Param.token3_index("0",")","=",j+1);
      if ( Param.equal(p-3,"(") && Param.equal(p - 4, var_id[i]) ) {
        (*var)[i] = KinExpression(Param, p+1, 0).Evaluate();
        break;
      }
    }
  }

  for (i = 0; i < ident_num; i++)
    ident_id[i] = Param.StrCpy( Param.token_index(IDENT_LABEL, i+1) - 1 );

  if (VERBOSE && ident_num)
    fprintf(stderr, "\n### Setting up %d time-dependent (auxiliary) variable(s):\n", ident_num);

  //if (VERBOSE && ident_num) fprintf(stderr, "\n### Sorting:\n");

  int ii = ident_num - 1;

  while ( ii > 0 )
      for (int j = 0; j < ii; j++)
        {
        int p, p0 = Param.token2_index(ident_id[j], IDENT_LABEL);
        int n = Param.tokens_to_eol(p0 + 1);
        for (p = p0 + 1; p <= p0 + n; p ++)
   	  if (equal(ident_id[ii], Param[p]) )
	     {
	     char *temp = ident_id[ii];
             ident_id[ii] = ident_id[j];
             ident_id[j] = temp;
             break;
	     }
        if ( p > p0 + n ) ii --;   // made through the whole loop without problem
        }

  for (i = 0; i < ident_num; i++)
    {
    if (VERBOSE) fprintf(stderr," #%d: %s == ", i+1, ident_id[i] );
    int p0 = Param.token2_index(ident_id[i], IDENT_LABEL) + 1;
    formulas[i] = new KinExpression(Param, p0, &Time, Ca, Bufs, this);
    ident->elem[i] = formulas[i]->Evaluate();
    if (VERBOSE) { formulas[i]->print(stderr,Ca, Bufs, this);
                   fprintf(stderr,"; %s(0) = %g\n",  ident_id[i], ident->elem[i] ); }
    }


  if (VERBOSE && var_num) fprintf(stderr, "\n### Setting up %d ODE(s):\n",var_num);

  for (i = 0; i < var_num; i++)
    {
    int p0;
    if (i < var_num1) p0 = Param.token_index(ODE_LABEL, i+1) + 1;
      else p0 = Param.token3_index("/","dt","=", i-var_num1+1) + 1;
    if (VERBOSE) fprintf(stderr," #%d: d%s/dt = ", i+1, var_id[i] );
    formulas[i + ident_num] = new KinExpression(Param, p0, &Time, Ca, Bufs, this);
    if (VERBOSE) { formulas[i + ident_num]->print(stderr, Ca, Bufs, this);
                   fprintf(stderr,"; %s(0) = %g\n", var_id[i], (*var)[i] ); } 
    }

  facilitation -> set_pointers(Param, Ca, Bufs, this);

  if ( Param.token_count("equilibrate") && var_num)
    {
      eq_flag = 1; 
      tables->Evaluate(Time);
      double eps = KINETIC_DEFAULT_EPS, eq_time = 1.0;
      long maxSteps = 40000;
      Param.trail_pars("equilibrate", 1, 'l', &maxSteps, 'd', &eps, 'd', &eq_time);
      if (VERBOSE) fprintf(stderr, 
        "\n### Equilibrating the ODE(s) (steps = %d, precision = %g, initial step = %g ms)",
         maxSteps, eps, eq_time);
      eq_time = RungeKuttaAdaptive(eq_time, eps, 0, 0.0, 0.0, maxSteps);
      if (VERBOSE) fprintf(stderr, "\n  # ODE(s) equilibrated after %g ms \n", eq_time);
      Time = 0.0;
    }

  facilitation -> reset();
  eq_flag = 0;
};

//**************************************************************************

void KineticObj::Evaluate()
 { 
 VectorObj dvar(var_num);

 if (!eq_flag)
   {
   locations->Evaluate(Time);
   tables->Evaluate(Time);
   }

 for (int i = 0; i < ident_num; i++)    ident->elem[i] = formulas[i]->Evaluate();
 facilitation -> Evaluate();
 }

//**************************************************************************

VectorObj KineticObj::derivative()
 { 
 VectorObj dvar(var_num);

 Evaluate();
 for (int i = 0; i < var_num;   i++)    dvar[i] = formulas[i+ident_num]->Evaluate();

 return dvar;
 }


//**************************************************************************
//      Cash-Karp embedded adaptive step-size Runge-Kutta scheme 
//                 (see Numerical Recipes, p. 710)
//**************************************************************************

double KineticObj::RungeKuttaAdaptive(double T, double eps, class PlotArray *plots, 
                                      double ceiling, double dt, long maxSteps) 
{
  long   count = 0;
  double NORMeps, norm;

  if ( eq_flag )                                       // 1st arg is initial equilibration time
    { dt = T; T = 1e20; NORMeps = eps; eps /= 10.0; }  // accuracy for the zero norm condition should be less 
                                                       // than the accuracy of the difference method
  else
    {
    if ( T <= 0 ) return Time;
    else if (dt <= 0) dt = T;    // dt is the initial time step, T is the integration time, ceiling is the 
    T += Time;                   //                                                  total simulation time
    if ( ceiling ) {
                   if ( Time >= ceiling )  return Time;
                   else if ( T > ceiling )  T = ceiling;
                   }
    if ( !var_num )  // if there are no ODEs to integrate, update all the time variables
      {
      Time = T;
      Evaluate();
      return Time;
      }
    }   // end if ( equilibrating )

  static double safety = 0.95;
  static const double a2 = 0.2, a3 = 0.3, a4 = 0.6, a5 = 1.0, a6 = 0.875;

  static const double b21 = 0.2;
  static const double b31 = 3.0 / 40.0, b32 = 9.0/40.0;
  static const double b41 = 0.3, b42 = -0.9, b43 = 1.2;
  static const double b51 = -11.0/54.0, b52 = 2.5, b53 = -70.0/27.0, b54 = 35.0/27.0;
  static const double b61 = 1631.0/55296.0, b62 = 175.0/512.0, b63 = 575.0/13824.0, b64 = 44275.0/110592.0, 
               b65 = 253.0/4096.0;

  static const double c1  = 37.0 / 378.0,     c3 = 250.0 / 621.0, c4 = 125.0 / 594.0, c6 = 512.0 / 1771.0;
  static const double cc1 = 2825.0 / 27648.0, cc3 = 18575.0 / 48384.0, cc4 = 13525.0 / 55296.0, cc5 = 277.0/14336.0,
               cc6 = 0.25;

 VectorObj var_old(*var), v1(var_num), v2(var_num); 
 VectorObj k1(var_num), k2(var_num), k3(var_num), k4(var_num), k5(var_num), k6(var_num);

 double delta;
 char   sdelta[80];

 do
   {
   if ( Time + dt > T )  dt = T - Time;
   double t0 = Time;

   k1 = derivative();
   *var = var_old + dt * (b21 * k1 );
   Time = t0 + a2 * dt;
   k2 = derivative();
   *var = var_old + dt * (b31 * k1 + b32 * k2 );
   Time = t0 + a3 * dt;
   k3 = derivative();
   *var = var_old + dt * (b41 * k1 + b42 * k2 + b43 * k3 ); 
   Time = t0 + a4 * dt;
   k4 = derivative();
   *var = var_old + dt * (b51 * k1 + b52 * k2 + b53 * k3 + b54 * k4 ); 
   Time = t0 + a5 * dt;
   k5 = derivative();
   *var = var_old + dt * (b61 * k1 + b62 * k2 + b63 * k3 + b64 * k4 + b65 * k5 ); 
   Time = t0 + a6 * dt;
   k6 = derivative();   

   v1 = c1  * k1 +  c3 * k3 +  c4 * k4            +  c6 * k6;
   v2 = cc1 * k1 + cc3 * k3 + cc4 * k4 + cc5 * k5 + cc6 * k6;

   delta = dt * norm_diff(v1, v2);
   sprintf(sdelta, "%g", delta);

   if (delta > eps || equal(sdelta,"nan") ) 
     {
     *var = var_old;
     Time = t0;
     if (VERBOSE > 7) fprintf(stderr, "# RK step: delta=%.3e > eps=%.3e time=%g dt=%.3g->", delta, eps, Time, dt);

     if (equal(sdelta,"nan")) dt /= 2.0; 
     else dt = safety * dt * pow(eps / delta, 0.25);
     if ( dt < DT_MIN ) {
       fprintf(stderr, "\n# ERROR: Runge-Kutta step: dt=%.3g reached the lower bound of %.3g", dt , DT_MIN);
       fprintf(stderr, "\n#      > delta=%.3e eps=%.3e time=%g\n", delta, eps, Time);
       fprintf(stderr, "\n#      > Decrease the accuracy parameters in \"Run\" definitions\n");
       // exit(1);
     }

     if (VERBOSE > 7) fprintf(stderr, "%.3g\n", dt);
       else if (VERBOSE > 2) { fprintf(stderr, "."); fflush(stderr); }
     continue;
     }
   else   // accuracy eps is satisfied
     {
     Time = t0 + dt;
     *var = var_old = var_old + v1 * dt;
     Evaluate();   // ensure that all variables are evaluated at the current time 
     if (plots) plots->draw_all();
     if (VERBOSE > 7) fprintf(stderr, "# RK step: delta=%.3e < eps=%.3e time=%g dt=%.3g->", delta, eps, Time, dt);
     if (delta > 0) dt = safety * dt * pow(eps / delta, 0.2); else dt = 2 * dt;
     if (VERBOSE > 7) fprintf(stderr, "%g\n", dt); 
       else if (VERBOSE > 2) { fprintf(stderr, "."); fflush(stderr); }
      }

   if ( eq_flag ) if ( (norm = derivative().norm() ) < NORMeps || ++count > maxSteps ) break;
   }
 while ( (T - Time) / (Time + T + 1e-12) > 1e-12 );

   if (count > maxSteps) 
     {
     fprintf(stderr, 
        "\n@@@ KineticObj equilibration: more than %ld iterations (norm = %.3g, step = %.3g ms) %c\n", 
        maxSteps, norm, dt, 7);
     fprintf(stderr, "@@@ Change the equilibration step or equilibration accuracy\n");
     fprintf(stderr, "@@@ Equilibration may not work if ODEs depend on time explicitly\n");
     fprintf(stderr, "    or if the ODE system is stiff\n");
     // exit(1);
     }

 if (VERBOSE > 7) fprintf(stderr, "## RK run completed: dt=%.3g\n", dt);

 return Time;
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

