#include "global.h"

extern int DIMENSIONALITY;

//*****************************************************************************
//                    G R I D   O P E R A T I O N S
//*****************************************************************************
   
 void set_dimensions(int nx, int ny, int nz)
 {
 nx = int(nx + 0.5);  nx |= 1;
 ny = int(ny + 0.5);  ny |= 1;
 nz = int(nz + 0.5);  nz |= 1;

 if (VERBOSE) { 
   fprintf(stderr, "\n### Setting up the grid: [0..%d]", nx - 1);
   if (DIMENSIONALITY > 1) fprintf(stderr, " x [0..%d]", ny - 1);
   if (DIMENSIONALITY > 2) fprintf(stderr, " x [0..%d]", nz - 1);
   fprintf(stderr, "\n");
 }

 xsize = nx; ysize = ny; zsize = nz;

 xysize = nx * ny;
 RegionObj *reg = FieldObj::get_region();
 Size = nx * ny * nz;

 xgrid = new double[nx]; xcoord = new double[nx];
 ygrid = new double[ny]; ycoord = new double[ny];
 zgrid = new double[nz]; zcoord = new double[nz];

 dxplus = new double[nx]; dxminus = new double[nx]; dxcenter = new double[nx];
 dyplus = new double[ny]; dyminus = new double[ny]; dycenter = new double[ny];
 dzplus = new double[nz]; dzminus = new double[nz]; dzcenter = new double[nz];
 dxinv = new double[nx];  dyinv  = new double[ny];  dzinv   = new double[nz];
 dx2inv = new double[nx]; dy2inv  = new double[ny]; dz2inv   = new double[nz];

 double dx = 1.0; double dy = 1.0; double dz = 1.0;

 if (nx > 1) dx = (reg->get_xmax() - reg->get_xmin()) / double(nx - 1);
 if (ny > 1) dy = (reg->get_ymax() - reg->get_ymin()) / double(ny - 1);
 if (nz > 1) dz = (reg->get_zmax() - reg->get_zmin()) / double(nz - 1);

 int i;
 for (i = 0; i < nx; i++)  
    { xgrid[i] = dx; xcoord[i] = reg->get_xmin() + i * dx; }
 for (i = 0; i < ny; i++)  
    { ygrid[i] = dy; ycoord[i] = reg->get_ymin() + i * dy; }
 for (i = 0; i < nz; i++)  
    { zgrid[i] = dz; zcoord[i] = reg->get_zmin() + i * dz; }

 int max = ( nx > ny ? nx : ny );
 if (nz > max) max = nz;
 init_tridiag(max + 2);
 }

//*****************************************************************************
   
 void set_dimensions(TokenString &params)
 {
 int nx=0, ny=0, nz=0;

 if (DIMENSIONALITY == 1)
   nx = params.trail_int("grid");
 else  if (DIMENSIONALITY == 2)
   params.trail_pars("grid", 1, 'i', &nx, 'i', &ny );
 else
   params.trail_pars("grid", 1, 'i', &nx, 'i', &ny, 'i', &nz);

 set_dimensions(nx, ny, nz);
 }

//*****************************************************************************

int location_to_index(double *coord, double *grid, int size, double x)
{
int i = 0;

while (x > coord[i] && i < (size - 1) ) i++;
if (i > 0) if (x < coord[i-1] + 0.5 * grid[i-1]) i --;

return i;
}

//*****************************************************************************

long location_to_index(double x, double y, double z)
{
int ix = location_to_index(xcoord, xgrid, xsize, x);
int iy = location_to_index(ycoord, ygrid, ysize, y);
int iz = location_to_index(zcoord, zgrid, zsize, z);

return ix + xsize * iy + xysize * iz;
}

//*****************************************************************************

void split(long ind, int &ix, int &iy, int &iz)
    {
    iz = ind / xysize;
    iy = (ind -= iz * xysize) / xsize;
    ix = ind - iy * xsize; 
    }

//*****************************************************************************

void grid_stretch(double *grid, int lim, double factor, int i0, int i1)
  {

  if (i0 > lim || i1 > lim || i0 < 0 || i1 < 0) 
    printf("global.cc::grid_stretch(): i0=%d i1=%d N=%d --> out of bounds\n",i0,i1,lim);

  if (i1 > i0) 
    {
    if (i0 == 0) i0 ++;
    for (int i = i0; i < i1; i++) grid[i] = grid[i-1] * factor;
    }

  if (i0 > i1) 
    {
    if (i0 > lim - 1) i0 = lim - 1;
    for (int i = i0 - 1; i >= i1; i--) grid[i] = grid[i+1] * factor;
    }
  }

/*****************************************************************************

 Non-uniform grid in [xmin, x1] and [x2, xmax];
 Stretch intervals [0, n1] and [n2, n]; 
 dx = unstretched grid distance within [n1, n2]:

  0             n1    n2                    N
  |    |   |  | ||||||| |  |   |     |      |
  xmin          x1    x2                    xmax
  <--   X1   --><- X -><--      X2        -->

 1) X = dx (n2 - n1)

                                         m=n1-1
                                           __
 2) x1 - xmin = dx f + dx f^2 + ... = f dx \   f^m = f dx (f^n1 - 1) / (f - 1) == X1
                                           /_  
                                           m=0

                                        m=N-n2-1
                                           __
 3) xmax - x2 = dx f + dx f^2 + ... = f dx \  f^m = f dx (f^(N-n2) - 1) / (f - 1) == X2
                                           /_  
                                           m=0

                  SOLUTION:
                  ^^^^^^^^
 I.   From (3): f^(N - n2) = [ X2 (f-1) / (f dx) ] + 1

 II.  From (1): f^n1 = f^(N - X/dx) / f^(N - n2) = f^(N - X/dx) / { [ X2 (f-1) / (f dx) ] + 1 }

 III. Substitute into (2):  X1 = { f^(N - X/dx) /  { [ X2 (f-1) / (f dx) ] + 1 } - 1 } f dx / (f - 1)

 IV.  Solve III for dx, using bisection

 V.   Get n1 from (2):  n1 = log(1 + X1 (f - 1) / (dx f) ) / log f

 VI.  Get n2 from (1):  n2 = X / dx + n1

*****************************************************************************/
 
void grid_stretch(char dir, double factor, double x1, double x2)
{
  int       N;
  double    *grid, *coord;
  double    eps = 1.0e-7, tiny = 1.0e-12;
 
  if (fabs(factor - 1) < 0.000001) return;

  switch (dir)
    {
    case 'x': grid = xgrid; coord = xcoord; N = xsize - 1; break;
    case 'y': grid = ygrid; coord = ycoord; N = ysize - 1; break;
    case 'z': grid = zgrid; coord = zcoord; N = zsize - 1; break;
    default: fprintf(stderr,"global.cc::grid_stretch: invalid axis label -- %c\n", dir); exit(1);
    }

  if (x1 < coord[0]) x1 = coord[0];
  if (x2 < coord[0]) x2 = coord[0];
  if (x1 > coord[N]) x1 = coord[N]; 
  if (x2 > coord[N]) x2 = coord[N]; 
  if (x2 < x1) 
    fprintf(stderr,"global.cc::grid_stretch: invalid uniform interval (%g > %g)\n",x1,x2), exit(1);

  if (VERBOSE) 
    fprintf(stderr,"  # Stretching grid with factor=%g along %c-axis leaving [%.4g,%.4g] uniform\n", 
            factor, dir, x1, x2);

  double X1 = x1 - coord[0];
  double X2 = coord[N] - x2;
  double X  = x2 - x1;
  double logf = log(factor);
  double f1f = (factor - 1) / factor;

  double dx0 = X / double(N) + tiny;                        // smallest possible dx
  double dx1 = (coord[N] - coord[0]) / double(N) - tiny;    // largest possible dx
  double f0 = X1 - dx0 / f1f * (exp( (N - X/dx0) * logf) / (1 + f1f * X2/dx0) - 1);
  double f1 = X1 - dx1 / f1f * (exp( (N - X/dx1) * logf) / (1 + f1f * X2/dx1) - 1);
  double dxn, fn;

  do     // Bisection begins
    {
    //printf("dx0=%f f0=%f   dx1=%f f1=%f\n",dx0,f0,dx1,f1);
    dxn = 0.5 * (dx0 + dx1);
    fn = X1 - dxn / f1f * (exp( (N - X/dxn) * logf ) / (1 + f1f * X2 / dxn) - 1);
    if (fn * f0 > 0) { dx0 = dxn; f0 = fn; }
      else { dx1 = dxn; f1 = fn; }
    }
  while (fabs( (dx0 - dx1) / (dx0 + dx1) ) > eps );

  int n1 = int( log(1 + X1 / dxn * f1f) / logf + 0.5 );   // step V.
  int n2 = int(X / dxn + n1 + 0.5 );                      // step VI.
  
  // if (n1 == n2) { n1--; n2++; }  ??? **** check if this is relevant

  if (n1 > 0) grid_stretch(grid, N, factor, n1, 0);
  if (n2 < N) grid_stretch(grid, N, factor, n2, N);

  int i = 0;              // RESCALE:
  double range = 0.0;

  for (i = 0; i < N; i++)  range += grid[i];
  double rescale = (coord[N] - coord[0]) / range;

  for (i = 0; i < N; i++)  grid[i] *= rescale; 
  for (i = 1; i < N; i++)  coord[i] = coord[i-1] + grid[i-1];
  grid[N] = grid[N-1];

  if (VERBOSE) {
    fprintf(stderr,"    Uniform %c-interval = [%.4g, %.4g] (%d..%d)\n", 
          dir, coord[n1], coord[n2], n1, n2);
    fprintf(stderr,"    Shortest dx = %.4g; dx on the boundaries: %.4g; %.4g\n", 
          grid[n1], grid[0], grid[N-1]);
  }
}

//*****************************************************************************

void grid_stretch(TokenString &params)
{
  char    dir;
  double  x1, x2;
  double  f = 1.0;
  int     n = params.token_count("stretch");

  if (params.token_count("stretch_factor") )
    {
    f = params.trail_double("stretch_factor");
    if (f < 1.0 || f >= 2.0) 
      fprintf(stderr, "global.cc::grid_stretch(): stretch factor = %g out of bounds\n",f), exit(1);
    }

  for (int i = 1; i <= n; i++)
    {
    params.trail_pars("stretch", i, 'c', &dir, 'd', &x1, 'd', &x2, 'E');  
    grid_stretch(dir, f, x1, x2);
    }
}
 

//*****************************************************************************

 void initScheme()
 { 
 int       i;

 if (GEOMETRY == 1) { initSchemeSpherical(); return; }
 else if (GEOMETRY == 2) { initSchemeSpherical(); initSchemeConical(); return; }

 for (i = 0; i < xsize; i++) 
   {
   dxplus[i]   = 2.0 / xgrid[i] / (xgrid[i] + xgrid[i ? i-1 : 0]);
   dxminus[i]  = 2.0 / xgrid[i ? i-1 : 0] / (xgrid[i] + xgrid[i ? i-1 : 0]);
   dxcenter[i] = -2.0 / (xgrid[i] * xgrid[i ? i-1 : 0]);
   dxinv[i] = 1.0 / xgrid[i];
   dx2inv[i] = 1.0 / (xgrid[i] * xgrid[i]);
   }

 for (i = 0; i < ysize; i++) 
   {
   dyplus[i]   = 2.0 / ygrid[i] / (ygrid[i] + ygrid[i ? i-1 : 0]);
   dyminus[i]  = 2.0 / ygrid[i ? i-1 : 0] / (ygrid[i] + ygrid[i ? i-1 : 0]);
   dycenter[i] = -2.0 / (ygrid[i] * ygrid[i ? i-1 : 0]);
   dyinv[i] = 1.0 / ygrid[i];
   dy2inv[i] = 1.0 / (ygrid[i] * ygrid[i]);
   }

 for (i = 0; i < zsize; i++) 
   {
   dzplus[i]   = 2.0 / zgrid[i] / (zgrid[i] + zgrid[i ? i-1 : 0]);
   dzminus[i]  = 2.0 / zgrid[i ? i-1 : 0] / (zgrid[i] + zgrid[i ? i-1 : 0]);
   dzcenter[i] = -2.0 / (zgrid[i] * zgrid[i ? i-1 : 0]);
   dzinv[i] = 1.0 / zgrid[i];
   dz2inv[i] = 1.0 / (zgrid[i] * zgrid[i]);
   }

 }

//*****************************************************************************

 void initSchemeSpherical()
 { 
 int       i;

 for (i = 0; i < xsize; i++)  {
   double r = xcoord[i];
   double drplus  = xgrid[i];
   double drminus = i ? xgrid[i-1] : xgrid[i];
   double rplus  = r + 0.5 * drplus;
   double rminus = r - 0.5 * drminus;
   double c = 2 / ( r * r * (drplus + drminus) );

   dxplus[i] = c * rplus * rplus / drplus;
   dxminus[i] = c * rminus * rminus / drminus;
   dxcenter[i] = - ( dxplus[i] + dxminus[i] );

   dxinv[i] = 1.0 / drplus;
   dx2inv[i] = 1.0 / (drplus * drplus);
   }

 }

//*****************************************************************************

 void initSchemeConical()
 { 
 int       i;

 for (i = 0; i < ysize; i++)  {
   double theta = ycoord[i];
   double dplus  = ygrid[i];
   double dminus = i ? ygrid[i-1] : ygrid[i];
   double sinplus  = sin(theta + 0.5 * dplus);
   double sinminus = sin(theta - 0.5 * dminus);
   double c = 2 / ( sin(theta) * (dplus + dminus) );

   dyplus[i] = c * sinplus / dplus;
   dyminus[i] = c * sinminus / dminus;
   dycenter[i] = - ( dyplus[i] + dyminus[i] );

   dyinv[i] = 1.0 / dplus;
   dy2inv[i] = 1.0 / (dplus * dplus);
   }
 }
//*****************************************************************************

 void kill_grid()
 { 
 if (VERBOSE > 7) { fprintf(stderr,"> Destructing grid data:"); fflush(stderr); }

 delete [] xgrid;  delete [] ygrid;   delete [] zgrid; 
 delete [] xcoord; delete [] ycoord;  delete [] zcoord;

 delete [] dxplus; delete [] dxminus; delete [] dxcenter;
 delete [] dyplus; delete [] dyminus; delete [] dycenter;
 delete [] dzplus; delete [] dzminus; delete [] dzcenter;

 delete [] dxinv;  delete [] dyinv;   delete [] dzinv;
 delete [] dx2inv; delete [] dy2inv;  delete [] dz2inv;

 if (VERBOSE > 7) fprintf(stderr," done\n");

 kill_tridiag();
 } 

//*****************************************************************************

void grid_clamp(TokenString &params)
{
  double  x, y, z;
  int     ix, iy, iz;
  int     n = params.token_count("clamp");

  for (int i = 1; i <= n; i++)
    {
    x = y = z = 0.0;
    params.trail_pars("clamp", i, 'd', &x, 'd', &y,  'd', &z,  'E');  

    if (  !in_bounds(x, y, z) )
      fprintf(stderr,"global.cc::grid_clamp(): position (%g,%g,%g) nm out of bounds\n", 
              x*1e3, y*1e3, z*1e3), exit(1);

    long ind = location_to_index(x, y, z);
    split(ind, ix, iy, iz);   

    if (VERBOSE) fprintf(stderr," > clamping x-node #%d: %.5g->%.5g nm; intervals: {%.5g,%.5g}", 
            ix, 1e3*xcoord[ix], 1e3*x, 1e3*(ix ? xgrid[ix-1] : 0 ), 1e3*xgrid[ix] );
    xcoord[ix] = x; 
    if (ix > 1) { xcoord[ix-1] = 0.5 * (xcoord[ix-2] + xcoord[ix]);
                  xgrid[ix-2] = xcoord[ix-1] - xcoord[ix-2]; }
    if (ix) xgrid[ix-1] = xcoord[ix] - xcoord[ix-1];
    if (ix < xsize-2) { xcoord[ix+1] = 0.5 * (xcoord[ix+2] + xcoord[ix]);
                          xgrid[ix+1]  = xcoord[ix+2] - xcoord[ix+1]; }
    if (ix < xsize-1) xgrid[ix] = xcoord[ix+1] - xcoord[ix]; else xgrid[ix] = xgrid[ix-1];
    if (VERBOSE) fprintf(stderr,"->{%.5g,%.5g} nm\n", 1e3*(ix ? xgrid[ix-1] : 0 ), 1e3*xgrid[ix] );

    
    if (VERBOSE) fprintf(stderr," > clamping y-node #%d: %.5g->%.5g nm; intervals: {%.5g,%.5g}", 
            iy, 1e3*ycoord[iy],  1e3*y, 1e3*(iy ? ygrid[iy-1] : 0 ), 1e3*ygrid[iy] );
    ycoord[iy] = y; 
    if (iy > 1) { ycoord[iy-1] = 0.5 * (ycoord[iy-2] + ycoord[iy]);
                  ygrid[iy-2] = ycoord[iy-1] - ycoord[iy-2]; }
    if (iy) ygrid[iy-1] = ycoord[iy] - ycoord[iy-1];
    if (iy < ysize-2) { ycoord[iy+1] = 0.5 * (ycoord[iy+2] + ycoord[iy]);
                          ygrid[iy+1]  = ycoord[iy+2] - ycoord[iy+1]; }
    if (iy < ysize-1) ygrid[iy] = ycoord[iy+1] - ycoord[iy]; else ygrid[iy] = ygrid[iy-1];
    if (VERBOSE) fprintf(stderr,"->{%.5g,%.5g} nm\n", 1e3*(iy ? ygrid[iy-1] : 0 ), 1e3*ygrid[iy] );
    

    if (VERBOSE) fprintf(stderr," > clamping z-node #%d: %.5g->%.5g nm; intervals: {%.5g,%.5g}", 
            iz, 1e3*zcoord[iz],  1e3*z, 1e3*(iz ? zgrid[iz-1] : 0 ), 1e3*zgrid[iz] );
    zcoord[iz] = z; 
    if (iz > 1) { zcoord[iz-1] = 0.5 * (zcoord[iz-2] + zcoord[iz]);
                  zgrid[iz-2] = zcoord[iz-1] - zcoord[iz-2]; }
    if (iz) zgrid[iz-1] = zcoord[iz] - zcoord[iz-1];
    if (iz < zsize-2) { zcoord[iz+1] = 0.5 * (zcoord[iz+2] + zcoord[iz]);
                          zgrid[iz+1]  = zcoord[iz+2] - zcoord[iz+1]; }
    if (iz < zsize-1) zgrid[iz] = zcoord[iz+1] - zcoord[iz]; else zgrid[iz] = zgrid[iz-1];
    if (VERBOSE) fprintf(stderr,"->{%.5g,%.5g} nm\n", 1e3*(iz ? zgrid[iz-1] : 0 ), 1e3*zgrid[iz] );
    
    }
}

//*****************************************************************************
//     B O U N D A R Y   C O N D I T I O N   D A T A   S T R U C T U R E S
//*****************************************************************************

   void  set_bc_types(TokenString &params)
   {
   int predefined = 2;

   bc_type_num = params.token_count("bc_type") + predefined;

   bc_deriv = new double[bc_type_num]; bc_lin = new double[bc_type_num];
   bc_coef  = new double[bc_type_num]; bc_id = new char *[bc_type_num];

   char s[256];
   double lin, deriv, coef;

   set_bc_type(0.0, 1.0, 0.0, "Dirichlet");
   set_bc_type(1.0, 0.0, 0.0, "Neumann");

   for (int i = 0; i < bc_type_num - predefined; i++)
     {
     params.trail_pars("bc_type", i+1, 's', s, 'd', &deriv, 'd', &lin, 'd', &coef);
     set_bc_type(deriv, lin, coef, s);
     if (VERBOSE) {
       fprintf(stderr, "\n### Defining boundary condition %s: ",bc_id[i+predefined]);
       if (bc_deriv[i+predefined]) fprintf(stderr, "du/dn");
       if (bc_lin[i+predefined])   fprintf(stderr, "%+g*u", bc_lin[i+predefined]);
       fprintf(stderr, " = %g \n", bc_coef[i+predefined]);
     }
     }
   }

//*****************************************************************************

   int  bcidtoint(const char *id)
   {
   for (int i=0; i<bc_type_num; i++)
     if ( strcmp(bc_id[i], id) == 0 ) return i ;
     
   printf("Error: bcidtoint: bc type %s does not exist\n",id);
   exit(1);
   }

//*****************************************************************************

   void  set_bc_types(int n)
   {
   bc_type_num = n;
   bc_deriv = new double[n]; bc_lin = new double[n]; bc_coef = new double[n];
   bc_id = new char *[n];
   }

//*****************************************************************************

   void  set_bc_type(double a, double b, double c, const char *id)
   {
   static int cnt = 0;

   if (id == 0) { cnt = 0; return; }  // reset counter

   if (cnt >= bc_type_num)
     {
     fprintf(stderr, "Error: set_bc_type: %d(cnt) >= %d (bc_type_num)\n",cnt,bc_type_num);
     exit(1);
     }

   if ( fabs(a) > 1e-10 ) 
     { b /= a; c /= a; a = 1; }  // normalization: if non-zero, derivative coefficient = 1
   else 
     if ( fabs(b) > 1e-10 ) { c /= b; b = 1; }  // if a = 0, then linear coefficient  = 1
     else
       { 
       fprintf(stderr, "Error: set_bc_type: values too small: a = %g, b = %g\n", a, b);
       exit(1);
       }
   
   bc_deriv[cnt] = a;  bc_lin[cnt] = b;  bc_coef[cnt] = c;

   bc_id[cnt] = StrCpy(id);
   cnt ++;
   }

//*****************************************************************************

   void  kill_bc()
   {
   if (VERBOSE > 7) { fprintf(stderr,"> Destructing Boundary condition definitions:"); fflush(stderr); }

   for (int i = 0; i < bc_type_num; i++) delete bc_id[i];

   delete [] bc_deriv; delete [] bc_lin; delete [] bc_coef;
   delete [] bc_id;

   set_bc_type(0, 0, 0, 0); // reset the static counter in set_bc_type

   if (VERBOSE > 7) fprintf(stderr," done\n");
   }


//*****************************************************************************

 char in_bounds(double x, double y, double z)
 {
    if (  x < xcoord[0] || x > xcoord[xsize-1] ||
          y < ycoord[0] || y > ycoord[ysize-1] ||
          z < zcoord[0] || z > zcoord[zsize-1] )
    return 0;
    else return 1;
 }

//*****************************************************************************
