#include "pde.h"
#include "global.h"

#define NORMAL 0
#define END    1

extern long Size;

extern int GEOMETRY;
extern const int SPHERICAL;
extern const int CONICAL;

//*************************************************************************
//*                T R I D I A G O N A L   S Y S T E M
//*************************************************************************


void init_tridiag(int n)
{
diag = new double[n];  right = new double[n];
sup = new double[n];   sub = new double[n];
}

void kill_tridiag()
{
  if (VERBOSE > 7) { fprintf(stderr,"> Destructing tridiagonal array:"); fflush(stderr); }
  delete [] diag; delete [] right;
  delete [] sup;  delete [] sub;
  if (VERBOSE > 7) fprintf(stderr," done\n");
}

//**************************************************************************
//                C A L C I U M   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ax/2 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay/2) Ca** = -Ay/2 Ca + Ca*
//  (1 - Az/2) Ca^  = -Az/2 Ca + Ca** + dt / 2 * { H(Ca^,B) - H(Ca,B) }
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int CaADIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       rvalue = NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(Size);

  if ( (T > 0.0) && (t0 + dt > T))  {  rvalue = END; dt = T - t0; }

  double nu = Ca.getD() * dt;
  Next = Ca;
  Lin = 0.0;
  double sum = 0.0;  
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= ( dt * Buf.array[b]->kminus) * *Buf.array[b];
    sum += dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.xRun(-0.5 * nu, 0.5 * nu,    nu,     nu,  0,  Next.elem);
  Ca.yRun(-0.5 * nu,     0.,  -0.5 * nu,  0.,  0,  Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Ca);

  Ca.zRun(-0.5 * nu, 0., 0., -0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//                  B U F F E R   D - G   S T E P
//**************************************************************************

int BufADIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( -dt * (Buf.kplus * Ca + Buf.kminus ) );
  Next = Buf % ( 1.0 + Lin ) + dt * Buf.kminus * Buf.total;

  Buf.xRun(-0.5 * nu, 0.5 * nu, nu, nu, 0, Next.elem);
  Buf.yRun(-0.5 * nu, 0., -0.5 * nu, 0., 0, Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Buf);
  Buf.zRun(-0.5 * nu, 0., 0., -0.5 * nu,  Lin.elem, Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next;  /*Buf.dirichlet();*/ }
  
  return rvalue;
}

//**************************************************************************
//                C A L C I U M   2 D   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ay/2) Ca  + dt/2 H(Ca,B)
//  (1 - Ay/2) Ca^  = (1 + Ax/2) Ca* + dt/2 H(Ca^,B)
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca2Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       b, rvalue = NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(Size);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  Lin = 0.0; 
  double sum = 0.0;  
  for (b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (0.5 * dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= *Buf.array[b] * (0.5 * dt * Buf.array[b]->kminus);
    sum += 0.5 * dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, 0.5 * dt);

  Ca.rRun(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem);

  for (b = 0; b < Buf.buf_num; b++)    
    Next -= *Buf.array[b] * (0.5 * dt * Buf.array[b]->kminus);
  Next += sum;
  Ca.add_sources(Next, 0.5 * dt);

  Ca.thetaRun(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.dirichlet();*/ }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//               B U F F E R   2 D   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) B*  = (1 + Ay/2) B  + dt/2 H(Ca,B)
//  (1 - Ay/2) B^  = (1 + Ax/2) B* + dt/2 H(Ca,B^)
//
//  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)
//
//  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf2Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( (-0.5 * dt) * (Buf.kplus * Ca + Buf.kminus) );

  Next = Buf + 0.5 * dt * Buf.kminus * Buf.total + Lin % Buf;

  Buf.rRun(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem);

  Next += 0.5 * dt * Buf.kminus * Buf.total;
  
  Buf.thetaRun(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next;  /*Buf.dirichlet();*/ }
  
  return rvalue;
}

//**************************************************************************
//                C A L C I U M   2 D   F I   S T E P
//**************************************************************************
//
//  (1 - Ax) Ca*  = (1 + Ay) Ca + dt H(Ca,B)
//  (1 - Ay) Ca^  = Ca* - Ay Ca + dt ( H(Ca^,B) - H(Ca,B) )
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) } + dt I
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca2DFIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       b, rvalue = NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(Size);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  Lin = 0.0; 
  double sum = 0.0;  
  for (b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= *Buf.array[b] * (dt * Buf.array[b]->kminus);
    sum += dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.rRun( -nu,  0.0,   nu,  0,  Next.elem);

  Next -= Lin % Ca;

  Ca.thetaRun(-nu, 0.0, -nu,  Lin.elem,  Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//               B U F F E R   2 D   F I   S T E P
//**************************************************************************
//
//  (1 - Ax) B*  = (1 + Ay) B + dt H(Ca,B)
//  (1 - Ay) B^  = B* - Ay B  + dt ( H(Ca,B^) - H(Ca,B) )
//
//  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)
//
//  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf2DFIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T))  { dt = T - Buf.Time; rvalue = END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( (-dt) * (Buf.kplus * Ca + Buf.kminus) );

  Next = Buf + dt * Buf.kminus * Buf.total + Lin % Buf;

  Buf.rRun(    -nu,     0.0,   nu,  0,  Next.elem);

  Next -= Lin % Buf;
  
  Buf.thetaRun(-nu, 0.0, -nu,  Lin.elem,  Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next; }
  
  return rvalue;
}

//**************************************************************************
//             C A L C I U M   L I N E A R   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ax/2 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay/2) Ca** = -Ay/2 Ca + Ca*
//  (1 - Az/2) Ca^  = -Az/2 Ca + Ca** + dt / 2 * { H(Ca^,B) - H(Ca,B) }
//
//  H(Ca,B) = Sum_B{ - kplus * (Ca * B0 + Ca0 * B) - kminus * B) }
//
//  Ca = Ca(t_n) - Ca0, Ca^ = Ca(t_{n+1}) - Ca0, B = B(t_{n+1/2}) - B0
//**************************************************************************

int CaLinADIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       b, rvalue = NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(Size);

  if ( (T > 0.0) && (t0 + dt > T))  {
    //printf("Ca: t0=%g dt=%g T=%g: dt reduced\n", t0,dt,T); 
    rvalue = END; 
    dt = T - t0; 
    }
  double nu = Ca.getD() * dt;

  Next = Ca - Ca.total;
  double cB, cCa, sum = 0;

  for (b = 0; b < Buf.buf_num; b++)
    {
    cB = dt * ( Buf.array[b]->kminus + Buf.array[b]->kplus * Ca.total );
    cCa = dt * Buf.array[b]->kplus * Buf.array[b]->bgr;
    Next -= *Buf.array[b] * cB + Ca * cCa;
    sum  += ( Buf.array[b]->bgr * cB + Ca.total * cCa );
    }

  Next += sum;

  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.xRun(-0.5 * nu, 0.5 * nu,    nu,     nu,  0,  Next.elem);
  Ca.yRun(-0.5 * nu,     0.,  -0.5 * nu,  0.,  0,  Next.elem);

  sum = 0.0; 
  for (b = 0; b < Buf.buf_num; b++)
    sum  += (0.5 * dt * Buf.array[b]->kplus) * Buf.array[b]->bgr;

  Lin = -sum;
  Next += sum * (Ca - Ca.total);

  Ca.zRun(-0.5 * nu, 0., 0., -0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next + Ca.total; Ca.Time += 0.5 * dt; /*Ca.dirichlet();*/ }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   L I N E A R   D - G   S T E P
//**************************************************************************

int BufLinADIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
VectorObj Lin(Size);
int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = END; }
  double nu = Buf.getD() * dt;
  
  double cB = dt * ( Buf.kminus + Buf.kplus * Ca.total );
  double cCa = dt * Buf.kplus * Buf.bgr;

  Next = Buf * (1 - cB) - Ca * cCa - ( Buf.bgr * (1 - cB) - Ca.total * cCa );

  Buf.xRun(-0.5 * nu, 0.5 * nu, nu, nu, 0, Next.elem);
  Buf.yRun(-0.5 * nu, 0., -0.5 * nu, 0., 0, Next.elem);

  Lin = -0.5 * cB;
  Next += 0.5 * cB * (Buf - Buf.bgr);
  Buf.zRun(-0.5 * nu, 0., 0., -0.5 * nu,  Lin.elem, Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next + Buf.bgr; /*Buf.dirichlet();*/ }

  return rvalue;
}

//**************************************************************************
//         C A L C I U M   F U L L Y   I M P L I C I T    S T E P
//**************************************************************************
//
//  (1 - Ax) Ca*  = (1 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay) Ca** = -Ay Ca + Ca*
//  (1 - Az) Ca^  = -Az Ca + Ca** + dt * { H(Ca^,B) - H(Ca,B) }  - no 1/2 !
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int CaFIADIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    b, rvalue = NORMAL;
  double t0 = Ca.Time;
  VectorObj Lin(Size); 
  
  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = END; }
  double nu = Ca.getD() * dt;  

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  double sum = 0.0;  
  for (b = 0; b < Buf.buf_num; b++)
    {
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca );
    sum += dt * Buf.array[b]->kminus * Buf.array[b]->total;
    }

  Next += sum;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.xRun(-nu, 0.0, nu, nu,  0,   Next.elem);
  Ca.yRun(-nu, 0.0, -nu,  0.,  0, Next.elem);

  Lin = 0.0; 
  for (b = 0; b < Buf.buf_num; b++)
  Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];  //<- no factor of 1/2 

  Next -= (Lin % Ca);
  Ca.zRun(-nu, 0., 0., -nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.dirichlet();*/ } else Ca.Time = t0;

  return rvalue;
}

//******************************************************************************************
//                B U F F E R   F U L L Y   I M P L I C I T    S T E P
//******************************************************************************************
//
//  (1 - Ax) Buf*  = (1 + Ay + Az) Buf + dt H(Buf,B)
//  (1 - Ay) Buf** = -Ay Buf + Buf*
//  (1 - Az) Buf^  = -Az Buf + Buf** + dt * { H(Buf^,Ca) - H(Buf,Ca) }  - no 1/2 !
//
//  H(Buf,B) = - kplus * Buf * Ca + kminus * (B_total - Buf) }
//
//  Buf = Buf(t_n), Buf^ = Buf(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int BufFIADIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ((1.0 - dt * Buf.kminus) - (dt * Buf.kplus) * Ca) + dt * Buf.kminus * Buf.total;

  Buf.xRun(-nu, 0.0, nu, nu, 0,  Next.elem);
  Buf.yRun(-nu, 0.0, -nu, 0., 0, Next.elem);

  VectorObj Lin( -dt * ( Buf.kplus * Ca + Buf.kminus ) );
  Next -= (Lin % Buf);
  Buf.zRun(-nu, 0.0, 0.0, -nu,  Lin.elem, Next.elem);
  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.dirichlet();*/ }

  return rvalue;
}

//**************************************************************************
//              C A L C I U M   E U L E R   S T E P
//**************************************************************************
//
//  Ca(t_{n+1}) = (1 + Ax + Ay + Az) Ca(t_n) + dt H(Ca(t_n), B(t_n))
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//**************************************************************************

int CaEulerStep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    rvalue = NORMAL;
  double t0 = Ca.Time;

  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = END; }
  double nu = Ca.getD() * dt;

  double sum = 0.0;  

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca );
    sum += dt * Buf.array[b]->kminus * Buf.array[b]->total;
    }
  Next += sum;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  Ca.zRun(0.0, nu, nu, nu, 0, Next.elem);
  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.dirichlet();*/ } else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   E U L E R   S T E P
//**************************************************************************

int BufEulerStep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ( 1.0 - dt * Buf.kminus - (dt * Buf.kplus) * Ca ) + dt * Buf.kminus * Buf.total;
 
  Buf.xRun(0.0, nu, nu, nu, 0, Next.elem);
  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.dirichlet();*/ }

  return rvalue;
}

//**************************************************************************
//              C A L C I U M   2 D   E U L E R   S T E P
//**************************************************************************
//
//  Ca(t_{n+1}) = (1 + Ax + Ay) Ca(t_n) + dt H(Ca(t_n), B(t_n))
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//**************************************************************************

int CaEu2Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    rvalue = NORMAL;
  double t0 = Ca.Time;

  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = END; }
  double nu = Ca.getD() * dt;

  double sum = 0.0;  

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca );
    sum += dt * Buf.array[b]->kminus * Buf.array[b]->total;
    }
  Next += sum;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  Ca.thetaRun(0.0, nu, nu, 0, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.dirichlet();*/ } else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   2 D   E U L E R   S T E P
//**************************************************************************

int BufEu2Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ( 1.0 - dt * Buf.kminus - (dt * Buf.kplus) * Ca ) + dt * Buf.kminus * Buf.total;
 
  Buf.thetaRun(0.0, nu, nu, 0, Next.elem);
  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.dirichlet();*/ }

  return rvalue;
}

/**************************************************************************
     C A L C I U M   C R A N K - N I C H O L S O N   1 D   S T E P
***************************************************************************

  (1 - Ax/2) Ca^ = (1 + Ax/2) Ca + dt/2 ( H(Ca, B) + H(Ca^,B) ) + dt sources(t_{n+1/2})
                 = (1 + Ax/2) Ca - dt/2 Sum_B{ kplus * B } ( Ca + Ca^ )
                 + dt { Sum_B{ kminus * (B_total - B) + sources(t_{n+1/2}) }

  ( 1 - Ax/2 - Lin ) Ca^ = Ax/2 Ca + (1 + Lin ) Ca
                         + dt { Sum_B{ kminus * (B_total - B) + sources(t_{n+1/2}) }

  Lin = - dt/2 Sum_B{ kplus * B }


  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }

  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
****************************************************************************/

int Ca1Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       rvalue = NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(Size);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  double sum = 0.0;  
  Lin = 0.0; 
  for (int b = 0; b < Buf.buf_num; b++) {
    Lin  -= (0.5 * dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= ( dt * Buf.array[b]->kminus ) * *Buf.array[b];
    sum += dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Ca % Lin;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.spheRun( -0.5 * nu, 0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.dirichlet();*/ }
    else Ca.Time = t0;

  return rvalue;
}

/**************************************************************************
       B U F F E R    C R A N K - N I C H O L S O N   1 D   S T E P
**************************************************************************

  (1 - Ax/2) B^ = (1 + Ax/2) B + dt/2 ( H(Ca, B) + H(Ca,B^) )
                = (1 + Ax/2) B - dt/2 kplus Ca ( B + B^ ) + dt kminus (B_total - B) 

  ( 1 - Ax/2 - Lin ) B^ = Ax/2 Ca + (1 + Lin ) B + dt kminus (B_total - B)

  Lin = - dt/2 kplus Ca

  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)

  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
****************************************************************************/

int Buf1Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = END; }
  double nu = Buf.getD() * dt;

  VectorObj Lin( (-0.5 * dt * Buf.kplus) * Ca ); 
  Next = Buf % ( 1.0 + Lin ) + (dt * Buf.kminus) * (Buf.total - Buf);

  Buf.spheRun( -0.5 * nu, 0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.dirichlet();*/ }

  return rvalue;
}
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

//**************************************************************************
//          H I N E S / N O N - A D A P T I V E    M E T H O D                 
//**************************************************************************

void FixedTimeStep(CaMethod CaStep, BufMethod BufStep, FieldObj &Ca, BufferArray &Buf, KineticObj &Gates,
            double T, int n, PlotArray &plots, double RKeps)
{
  VectorObj Next(Size);
  double dt = (T - Ca.Time) / double(n + 0.5);

  Ca.Step(CaStep, Next, Buf, 0.5 * dt, T);  // half a step for staggering
  
  for (int i = 1; i <= n; i++)
    {
    Buf.Step(BufStep, Next, Ca, dt, T);
    Gates.RungeKuttaAdaptive(dt, RKeps, 0, T);
    Ca.Step(CaStep, Next, Buf, dt, T);
    if (VERBOSE > 1) { fprintf(stderr, ","); fflush(stderr); }
    plots.draw_all(); 
    }

  Buf.Step(BufStep, Next, Ca, 0.5 * dt, T); // do half a time step for staggering
  Gates.RungeKuttaAdaptive(0.5 * dt, RKeps, 0, T);
  plots.draw_all();
}

//**************************************************************************
//                    F I N D   T I M E   S T E P
//**************************************************************************

void SetTimeStep(CaMethod CaStep, FieldObj &Ca, BufferArray &Buf,  double &dt, double eps)
{
  VectorObj Next(Size); 
  FieldObj  Next2(Ca); 
  double    old, half, full, error;

  if (VERBOSE) {
    fprintf(stderr, " ## Setting the initial time step ,"); fflush(stderr); }

  Ca.Step(CaStep, Next, Buf, dt, -1); 
  half = Next.norm();

    do
      {
      old = half;
      dt *= 0.5;
      Ca.Step(CaStep, Next2, Buf, dt, -1);
      half = Next2.norm();
      Next2.Time = Ca.Time + dt;
      Next2.Step(CaStep, Next, Buf, dt, -1);
      full = Next.norm();
      if (VERBOSE > 1) { fprintf(stderr, ","); fflush(stderr); }
      else if (VERBOSE > 5)
         fprintf(stderr, " > halving dt: dt = 2 * %g: res=%g  dt=%g: res=%g\n",dt,full,2*dt,old);
      }
    while ( (error = fabs( 2 * (old - full) / (old + full + 1.0e-12) )) > eps );
    Ca = Next2;
    // Ca.Time += dt;
    dt *= 2;
    if (VERBOSE) 
        fprintf(stderr, ", set to %.2e ms (error=%.2e) \n", dt,error);
}

//**************************************************************************
//                 A D A P T I V E   M E T H O D                 
//**************************************************************************
// Iterations begin with buffers update, so
// be sure to pass Ca field iterated over dt/2 !!
//**************************************************************************

void Adaptive(CaMethod CaStep, BufMethod BufStep, FieldObj &Ca, BufferArray &Buf, KineticObj &Gates,
	      double T, double dt, PlotArray &plots, double eps, double &dtmax,
              double &expand, int &n0, double &more, int &nmax)
{
  VectorObj Next(Size); 
  FieldObj  Next2(Ca), oldCa(Ca);
  BufferArray oldBuf(Buf);  
  double    one, two, buf_dt, oldDt;
  int       rvalue, total_steps = 0, between_divides = 0;
  double    error, max_error = 0.05, correct = 0.8;  
  double    dt_new;
  double    RKeps = eps;
 
  if (RKeps > 1e-3) RKeps = 1e-3;
  double between_checks = n0;

  while(1)
    {
    int i = 1;
    oldCa = Ca;        oldBuf = Buf;   // store variables to recover 
    Gates.saveState(); oldDt = dt;     // when max_error is exceeded

    do         // loop over n="between_checks" iterations
      { 
      if ( ( dt_new = dt * expand ) > dtmax ) dt_new = dtmax;
      buf_dt = 0.5 * ( dt + dt_new );
      Buf.Step(BufStep, Next, Ca, buf_dt, T); 
      Gates.RungeKuttaAdaptive(buf_dt, RKeps, 0, T);             
      dt = dt_new;                            
      rvalue = Ca.Step(CaStep, Next, Buf, dt, T);
      plots.draw_all();
      total_steps ++; 
      between_divides ++; 
      if (VERBOSE > 1) { fprintf(stderr, ","); fflush(stderr); }
      else if (VERBOSE > 5) 
        fprintf(stderr, "\n > Adaptive step #%d(of %d): T=%g(%g) dt=%.3g", 
              i, int(between_checks+0.5), Ca.Time, Gates.Time, dt);
      if ( (i++ == int(between_checks + 0.5)) && (Ca.Time + dt < T) ) break; 
      }
    while (!rvalue);

    Buf.Step(BufStep, Next, Ca, dt, T);
    Gates.RungeKuttaAdaptive(dt, RKeps, 0, T);
    plots.draw_all();
    if (rvalue) return;

    Ca.Step(   CaStep, Next2, Buf, 0.75 * dt, -1);  // check accuracy 
    Next2.Step(CaStep, Next,  Buf, 0.25 * dt, -1);
    two = Next.norm();
    Ca.Step(CaStep, Next, Buf, dt, -1);
    one = Next.norm();
    error = fabs( 2 * (two - one) / (two + one + 1.0e-12) );

    if (VERBOSE > 4) 
      fprintf(stderr,"\n > accuracy check after %d steps: T=%g dt=%g error=%g ", 
              between_divides, Ca.Time, dt, error);
       
    if (error > max_error) {
      more = 1 + (more - 1) * correct; 
      if (n0 > 1) n0--;
      if (nmax > 1) nmax--;
      expand = 1 + (expand - 1) * correct;
      fprintf(stderr,"\n*** WARNING:  max_error exceeded (error = %.2g%% > %.2g%%)\n", error*100, max_error*100);
      fprintf(stderr,"  > Stepping back by %.3gms and halving time step to %.3ems\n", Ca.Time-oldCa.Time, dt*0.5);
      fprintf(stderr,"  > Changing adaptive time-step parameters: \n");
      fprintf(stderr,"  > expansion=%g initial_steps=%d, step_increase=%g, max_steps=%d\n", expand, n0, more, nmax); 
      Ca = oldCa; Buf = oldBuf; Gates.recoverState();
      Buf.Step(BufStep, Next, Ca, oldDt * 0.75, T); 
      Gates.RungeKuttaAdaptive(oldDt * 0.75, RKeps, 0, T);             
      dt *= 0.5;
      rvalue = Ca.Step(CaStep, Next, Buf, 0.5 * dt, T);
      plots.draw_all();
      } 
    else if ( error > eps )  // halving the time step
       {
       Next2.Time = Ca.Time;
       Ca = Next2;
       Ca.Time += 0.75 * dt;
       dt /= 2; 
       between_checks = n0;
       if (VERBOSE > 2)
         fprintf(stderr, "\n# dt halved @ %d steps: t=%gms, dt=%.2ems, error=%.2e ", 
                between_divides, Ca.Time, dt, error);
       else fprintf(stderr, "|", dt);
       between_divides = 0;
       }
     else   // accuracy within limits (error < eps), time step unchanged
       {
       if (VERBOSE) fprintf(stderr, "+");
       if ( int(between_checks *= more) > nmax) between_checks = nmax;       
       Ca = Next;
       Ca.Time += dt;
       }
     }
}

//**************************************************************************
//                                R U N                
//**************************************************************************

void Run(TokenString &Params, FieldObj &Ca, BufferArray &Bufs, KineticObj &Gates,
         class PlotArray &plots)
{
  CaMethod CaStep = 0;
  BufMethod BufStep = 0;
  double t, T = 0.0, dt = 1.0e-3;
 
  static double eps1=1.0e-3, eps2=1.0e-4, dtmax = 0.1, gamma=1.05, more=1.2;
  int    n0 = 2, nmax = 22;

  plots.draw_all();

  for (int i = 1; i <= Params.token_count("Run"); i++)
        {
	bool adaptive = true;
        char smethod[100];

        Params.trail_string("Run", smethod, i);

        if ( equal(smethod, "Adaptive") ) {
          Params.trail_pars("Run", i, 0,0, 's', smethod, 'E');
          if ( Params.is_number(smethod) ) {
            strcpy(smethod, "default");
            Params.trail_pars("Run", i, 0,0, 'd', &t, 'd', &dt, 'd', &dtmax, 'd', 
                              &eps1, 'd', &eps2, 'd', &gamma, 'i', &n0, 'd', &more, 'i', &nmax);
	  } else
             Params.trail_pars("Run", i, 0,0, 0,0, 'd', &t, 'd', &dt, 'd', &dtmax, 
                               'd', &eps1, 'd', &eps2, 'd', &gamma, 'i', &n0, 'd', &more, 'i', &nmax);
        } else {
          adaptive = false;
          if ( Params.is_number(smethod) ) {
            strcpy(smethod, "default");
            Params.trail_pars("Run", i, 'd', &t, 'd', &dt, 'd', &eps2);
	  } else
             Params.trail_pars("Run", i, 's', smethod, 'd', &t, 'd', &dt, 'd', &eps2);
	}  

        KinExpression Current(Params, Params.token_index("current", i) + 1, &(Ca.Time),
                              &Ca, &Bufs, &Gates );
        Ca.current = &Current;

        T += t;

        getMethod(smethod, CaStep, BufStep );

        if ( adaptive )
	  { 
          fprintf(stderr,"\n\n#### Running PDEs using adaptive %s method for %g ms, current = ", 
                  smethod, t);
          Current.print(stderr, &Ca, &Bufs, &Gates );
          fprintf(stderr," [I]\n");
          fprintf(stderr, " > dt0=%gms, dt_max=%gms, initial_acc=%g, accuracy=%g, expansion=%g\n",
                            dt, dtmax, eps1, eps2, gamma);
          fprintf(stderr, " > initial_steps=%d, step_increase=%g, max_steps=%d\n",
                            n0, more, nmax); 

          if (dt > dtmax) dt = dtmax;
          SetTimeStep(CaStep, Ca, Bufs, dt, eps1);
          Adaptive(CaStep, BufStep, Ca, Bufs, Gates, T, dt, plots, 
                   eps2, dtmax, gamma, n0, more, nmax);
          dt *= 10;
	  }
        else   // non-adaptive method
	  {
          int Tsteps = int(t / dt + 0.5);
          fprintf(stderr,"\n#### Running PDEs using non-adaptive %s method for %g ms, current = ",
                  smethod, t);
          Current.print(stderr, &Ca, &Bufs, &Gates );
          fprintf(stderr," [I]\n > time step = %g ms, number of iterations = %d\n", dt, Tsteps);
          FixedTimeStep(CaStep, BufStep, Ca, Bufs, Gates, T, Tsteps, plots, eps2);
	  }
        plots.redraw_all();
	} // for loop over runs
  fprintf(stderr, "\n### Simulation completed: total time = %g ms\n",T );
}

//**************************************************************************

void getMethod( char *smethod, CaMethod &CaStep, BufMethod &BufStep) {

 switch (GEOMETRY)  { 
 case 0:                        // 3D cartesian geometry
     if ( equal(smethod,"1D") )
       { CaStep = &Ca1Dstep; BufStep = &Buf1Dstep; strcpy(smethod,"radially symmetric ADI"); }
     else if ( equal(smethod,"LIN") )
       { CaStep = &CaLinADIstep; BufStep = &BufLinADIstep; strcpy(smethod,"linearized ADI"); }
     else if ( equal(smethod,"FI") )
       { CaStep = &CaFIADIstep; BufStep = &BufFIADIstep;  strcpy(smethod,"fully implicit"); }
     else if ( equal(smethod,"EU") )
       { CaStep = &CaEulerStep; BufStep = &BufEulerStep; strcpy(smethod,"Euler"); }
     else if ( equal(smethod,"DG") || equal(smethod,"default") ) // default
       { CaStep = &CaADIstep; BufStep = &BufADIstep; strcpy(smethod,"Douglas-Gunn ADI"); }
     else  printf("pde.cc: method %s not recognized\n", smethod);
     break;

 case 1:                          // ** SPHERICAL
     if ( equal(smethod,"CN") || equal(smethod,"default") )
       { CaStep = &Ca1Dstep; BufStep = &Buf1Dstep; strcpy(smethod,"spherical 1D Crank-Nicholson implicit"); }
     else  printf("pde.cc: method %s not recognized\n", smethod);
     break;

 case 2:                          // *** CONICAL
     if ( equal(smethod,"DG") || equal(smethod,"default") )
       { CaStep = &Ca2Dstep; BufStep = &Buf2Dstep; strcpy(smethod,"conical (r,theta) 2D Crank-Nicholson implicit"); }
     else  if ( equal(smethod,"EU") )
       { CaStep = &CaEu2Dstep; BufStep = &BufEu2Dstep; strcpy(smethod,"conical (r,theta) 2D Euler"); }
     else  if ( equal(smethod,"FI") )
       { CaStep = &Ca2DFIstep; BufStep = &Buf2DFIstep; strcpy(smethod,"conical (r,theta) 2D fully implicit"); }
     else  printf("pde.cc: method %s not recognized\n", smethod);
     break;

 }

}

//**************************************************************************

double get_sim_time(TokenString &params, int &sim_steps)
  {
  char   ptype[256], s2[256];
  double dt, tt, t = 0.0;
  int    n = 0;

  if ( params.token2_count("Run","ODE"))
    {
    n = sim_steps;
    for (int i = params.token2_count("Run","ODE"); i >=1; i--)
      t += params.trail2_double("Run", "ODE", i);
    }
  else  
   for (int i = params.token_count("Run"); i >=1; i--)
     {
     params.trail_pars("Run", i, 's', ptype, 's', s2, 'E');
     if ( equal(ptype, "Adaptive") ) 
       {
       if ( params.is_number(s2) ) params.trail_pars("Run", i, 0,0, 'd', &tt, 'E');
         else params.trail_pars("Run", i, 0,0, 0,0, 'd', &tt, 'E');
       t += tt; n = sim_steps;
       } 
     else  
      {
      if ( params.is_number(ptype) ) tt = params.token_double(ptype);
      else params.trail_pars("Run", i, 0, 0, 'd', &tt, 'd', &dt, 'E');
      t += tt; n += int(tt / dt + 0.5) + 2;
      }
     }
     
  sim_steps = n;
  return t;
  }

//**************************************************************************
//                           O D E   S O L V E R               
//**************************************************************************

void RunODEs(TokenString &param,  KineticObj &Gates, PlotArray *plots)
{
  double time = 0;
  double dt = 1.0;
  double eps = 1.0e-6;
  int dummy;

  double total = get_sim_time(param, dummy);
  
  fprintf(stderr,"\n\n#### Running the ODEs: total simulation time = %g ms\n",total);

  for (int i = 1; i <= param.token2_count("Run","ODE"); i++)
    {
    param.trail2_pars("Run", "ODE", i, 'd', &time, 'd', &dt, 'd', &eps, 'E');

    fprintf(stderr," ### ODE Run #%d: time = %g ms, accuracy = %g, initial time-step = %g ms\n", 
            i, time, eps, dt);

    Gates.RungeKuttaAdaptive(time, eps, plots, total, dt);
    if (VERBOSE) fprintf(stderr,"\n");
    plots->redraw_all();
    }
}
//**************************************************************************

