#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "syntax.h"
#include "vector.h"
#include "box.h"
#include "field.h"
#include "expr.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "gate.h"


//***************************************************************************

TokenString::TokenString(const char *fname, const char *extra)
{
  FILE   *f;
  verbose = 0;

  if (!(f = fopen(fname,"r"))) 
    { fprintf(stderr,"\n*** ERROR: TokenString::TokenString:: no file %s\n", fname); exit(1); }

  carr_ret = ::StrCpy(CARR_RET_TOKEN);

  storage   = new char[max_length = MAX_STORAGE_LENGTH];
  token_ptr = new char *[max_token_num = MAX_TOKEN_NUM];

  length = 0;
  token_num = 0;
  token_ptr[0] = storage;

  parse(fname, extra);
}

//***************************************************************************

void TokenString::parse(const char *fname, const char *extra) {
  FILE   *f;
  bool exprFlag = false;   // in expression, minus/plus signs are treated as separate tokens

  if (!(f = fopen(fname,"r"))) 
    { fprintf(stderr,"\n*** ERROR: TokenString::parse: no file %s\n", fname); exit(1); }

  char s[1024];  // max_string_length
  
  while (1)   {
    if ( (token_num == 0) && extra ) strcpy(s, extra);
    else { fgets(s,MAX_STORAGE_LENGTH,f); if (feof(f)) break; }

    char *ptr0=s, *ptr1, *last = s + strlen(s);
    int tnum_old = token_num;
    
    do { 
      getToken(ptr0,ptr1,last, exprFlag);
      if ( ::equal_(ptr0,ODE_LABEL) || ::equal_(ptr0,IDENT_LABEL) ||
           ::equal_(ptr0,"=") || ::equal_(ptr0,"current") ) exprFlag = true;  // is an expression

      if ( ::equal_(ptr0,COMMENT_TOKEN) || ::equal_(ptr0,LINE_CONTINUE_TOKEN) )  break;

      if ( ::equal_(ptr0,INPUT_TOKEN) ) {     // import data file
        if ( (ptr0 = ptr1) >= last ) break;
        getToken(ptr0,ptr1,last, exprFlag);
        char *input_file = ::StrCpy(ptr0,ptr1-ptr0);
        parse(input_file);
        delete [] input_file;
        if ( (ptr0 = ptr1) >= last ) break; 
          else continue;
      }
      addToken(ptr0,ptr1 - ptr0);
      ptr0 = ptr1;
    } while (ptr1 < last );

    if ( !::equal_(ptr0,LINE_CONTINUE_TOKEN) && token_num > tnum_old ) {
      addToken(CARR_RET_TOKEN, strlen(CARR_RET_TOKEN) );
      exprFlag = false;   // by default each new line is not an expression
    }
  }

  fclose(f);
}

//***************************************************************************

void TokenString::getToken(char *&ptr0, char *&ptr1, char *last, bool expression_flag)
{
  ptr1 = ptr0;

  while ( is_delim( *ptr0 ) && ( ptr0 < last ) )  ptr1 = ++ptr0;
  if (ptr1 == last) return;

  if (*ptr0 == '\"' || *ptr0 == '\'' || *ptr0 == '`') {
    char quote = *ptr0;
    ptr1 = ++ptr0;
    while ( *(++ptr1) != quote ) 
      if (ptr1 >= last) { 
        fprintf(stderr,"\n*** ERROR: unterminated string in parameter file:\n %s\n",ptr0);
        exit(1);
      }
    *ptr1 = ' ';  // erase closing quote
    return;
  }

  if ( ::equal_(ptr0, ODE_LABEL) ) { ptr1 += strlen(ODE_LABEL); return; }
  if ( ::equal_(ptr0, IDENT_LABEL) ) { ptr1 += strlen(IDENT_LABEL); return; }

  if ( ::equal_(ptr0, LINE_CONTINUE_TOKEN) )  { ptr1 += strlen(LINE_CONTINUE_TOKEN); return; }
  if ( ::equal_(ptr0, CARR_RET_TOKEN) )  { ptr1 += strlen(CARR_RET_TOKEN); return; }
  if ( ::equal_(ptr0, COMMENT_TOKEN) )  { ptr1 += strlen(COMMENT_TOKEN); return; }

  if ( is_short_token(*ptr1) && !is_sign(*ptr1) ) { ptr1++; return; }    // single-character token

  if ( is_sign(*ptr1) ) {         // a sign is a separate token only in an expression
    ptr1++;
    if (expression_flag) return;
  }

  if ( is_numeral(*ptr1) || (*ptr1 == '.' && is_numeral(ptr1[1]) ) ) {    // getting a number
    while ( is_numeral(*ptr1) ) ptr1++;
    if (*ptr1 == '.') ptr1++;
    while ( is_numeral(*ptr1) ) ptr1++;
    if ( (*ptr1 == 'e' || *ptr1 == 'E') && 
         ( is_numeral(ptr1[1]) || ( is_sign(ptr1[1]) && is_numeral(ptr1[2]) ) ) ) {  // getting exponent
      ptr1++;
      if ( is_sign(*ptr1) ) ptr1++;
      while ( is_numeral(*ptr1) ) ptr1++;
    }
    else                                                          // otherwise it's a weird identifier  
      while (!is_short_token(*ptr1) && !is_delim(*ptr1)) ptr1 ++; // ("plot 2D", for instance)
    return;
  }

  while (!is_short_token(*ptr1) && !is_delim(*ptr1))  // some sort of an identifier
    ptr1 ++;

  return;
}

//***************************************************************************

void TokenString::addToken(const char *token, int tlength)
{
  if (tlength == 0) return;

  if (length + tlength + 1 >= max_length)
    {
    char *buffer = new char[max_length += STORAGE_INCREASE];   
    for (long l = 0; l < length; l++) buffer[l] = storage[l];
    for (int i = 0; i <= token_num; i++) token_ptr[i] = buffer + (token_ptr[i] - storage);
    delete [] storage;
    storage = buffer;
    }

  if (token_num + 1 >= max_token_num)
    {
    char **tptrs = new char *[max_token_num += TOKEN_NUM_INCREASE];   
    for (int i = 0; i <= token_num; i++) tptrs[i] = token_ptr[i];
    delete [] token_ptr;
    token_ptr = tptrs;
    }

  strncpy( token_ptr[token_num], token, tlength );
  *(token_ptr[token_num] + tlength) = 0;

  //fprintf(stderr, "%s ", token_ptr[token_num]);

  token_ptr[token_num+1] = token_ptr[token_num] + tlength + 1;
  token_num++;
  length += tlength + 1;
}

//***************************************************************************

TokenString::TokenString(const TokenString &tstring)
{
  token_num = tstring.token_num;
  length    = tstring.length;

  storage = new char[length];
  token_ptr = new char *[token_num + 1];
 
}

//***************************************************************************

int TokenString::token_count(const char *token)
{
  int cnt = 0;
  for (int i=0; i<token_num; i++)  if ( equal(i, token) ) cnt ++;
  return cnt;
}

//***************************************************************************

int TokenString::token2_count(const char *token1, const char *token2 )
{
  int cnt = 0;
  for (int i = 1; i < token_num; i++)
   if ( equal(i-1, token1) && equal(i, token2) ) cnt ++;
  return cnt;
}

//***************************************************************************

int TokenString::token3_count(const char *token1, const char *token2, const char *token3 )
{
  int cnt = 0;
  for (int i = 2; i < token_num; i++)
   if ( equal(i-2, token1) && equal(i-1, token2) && equal(i, token3) ) cnt ++;
  return cnt;
}
//***************************************************************************

int TokenString::token_count_(const char *token)
{
  int cnt = 0;
  for (int i=0; i<token_num; i++)  if ( equal_(i, token) ) cnt ++;
  return cnt;
}

//***************************************************************************

void TokenString::print(const char *fname)
{
  FILE *f;

  if ( !(f = fopen(fname, "w")) ) 
    fprintf(stderr,"\nTokenString::print: could not open %s\n", fname);

  fprintf(f, "# ");
  for (int i = 0; i < token_num; i++)
   fprintf(f, "%s ", token_ptr[i]);
  fprintf(f, "\n");
}

//***************************************************************************

void TokenString::print(FILE *f)
{

  fprintf(f, "# ");
  for (int i = 0; i < token_num; i++)
   fprintf(f, "%s ", token_ptr[i]);
  fprintf(f, "\n");
}

//***************************************************************************

void TokenString::trail_pars(const char *token, int m, ... )
{
char    type;
int    *ipar;
double *dpar;
long   *lpar;
char   *cpar;
char   *spar;

 int cnt = 0, ind = 0;
  if (verbose) fprintf(stderr, "Parser: %s --> \n",token);

  do  {
    if ( equal(token, ind) ) cnt ++;
    if (cnt == m) break;
    }    
  while ( ++ind < token_num );

  if (cnt < m)  {
    fprintf(stderr,"\nTokenString::trail_pars: less than %d instances of %s\n",m,token);
    exit(1);
    }

  va_list p;
  va_start(p, m);

  while (1)
    {
    ind ++; 
    if ( equal(ind, carr_ret) ) break;
    type = char( va_arg(p, int) );
    if (type == 'E' ) break;

    switch (type)
      {
      case 'd': dpar =  va_arg(p, double *);
 	        *dpar = index_double(ind);
                if (verbose) fprintf(stderr,"  --> %g \n", *dpar); break;
      case 'i': ipar =  va_arg(p, int *);   
                *ipar = index_int(ind);
                if (verbose) fprintf(stderr,"  --> %d \n", *ipar); break;
      case 'l': lpar =  va_arg(p, long *);
                *lpar = index_long(ind);
                if (verbose) fprintf(stderr,"  --> %ld \n", *lpar); break;
      case 'c': cpar =  va_arg(p, char *);   sscanf(token_ptr[ind],"%c",cpar); 
                if (verbose) fprintf(stderr,"  --> %c \n", *cpar); break;
      case 's': spar =  va_arg(p, char *);   
                index_string(ind,spar); 
                if (verbose) fprintf(stderr,"  --> %s \n", spar); break;
      default:  va_arg(p, int); break;  // ignore argument 
      }
    }
  va_end(p);
}

//***************************************************************************

void TokenString::trail2_pars(const char *token1, const char *token2, int m, ... )
{
char    type;
int    *ipar;
double *dpar;
long   *lpar;
char   *cpar;
char   *spar;

  int cnt = 0, ind = 1;
  if (verbose) fprintf(stderr, "Parser: %s %s -->\n",token1, token2);

  do  {
    if ( equal(token1, ind-1) && equal(token2, ind) ) cnt ++;
    if (cnt == m) break;
    }    
  while ( ++ind < token_num );

  if (cnt < m)  {
    fprintf(stderr,"\nTokenString::trail_params2: less than %d instances of %s %s\n",
            m, token1, token2);
    exit(1);
    }

  va_list p;
  va_start(p, m);

  while (1)
    {
    ind ++;  
    type = char( va_arg(p, int) ); 
    if ( type == 'E' )  break; 
    if ( equal(ind, carr_ret) ) break;
    switch (type)
      {
      case 'd': dpar =  va_arg(p, double *);
 	        *dpar = index_double(ind);
                if (verbose) fprintf(stderr,"  --> %g \n", *dpar); break;
      case 'i': ipar =  va_arg(p, int *);   
                *ipar = index_int(ind);
                if (verbose) fprintf(stderr,"  --> %d \n", *ipar); break;
      case 'l': lpar =  va_arg(p, long *);
                *lpar = index_long(ind);
                if (verbose) fprintf(stderr,"  --> %ld \n", *lpar); break;
      case 'c': cpar =  va_arg(p, char *);   sscanf(token_ptr[ind],"%c",cpar); 
                if (verbose) fprintf(stderr,"  --> %c \n", *cpar); break;
      case 's': spar =  va_arg(p, char *);
                index_string(ind,spar); 
                if (verbose) fprintf(stderr,"  --> %s \n", spar); break;
      default:  va_arg(p, int); break;  // ignore argument 
      }
    }
  va_end(p);
}

//***************************************************************************

int TokenString::token_index(const char *token, int n) 
     { 
     int cnt = 0;
     for (int i = 0; i < token_num; i++)
       {
       if ( equal(i, token) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token_index: less than %d instances of [%s]\n",n,token),
     exit(1);
     }

int TokenString::token2_index(const char *token1, const char *token2, int n)
     { 
     int cnt = 0;
     for (int i = 1; i < token_num; i++)
       {
       if ( equal(i-1, token1) && equal(i, token2) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token2_index: less than %d instances of [%s %s]\n",n,token1,token2),
     exit(1);
     }

int TokenString::token3_index(const char *token1, const char *token2, const char *token3, int n)
     { 
     int cnt = 0;
     for (int i = 2; i < token_num; i++)
       {
       if ( equal(i-2, token1) && equal(i-1, token2) && equal(i, token3) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token3_index: less than %d instances of [%s %s]\n",n,token1,token2),
     exit(1);
     }

int TokenString::token_index_(const char *token, int n) 
     { 
     int cnt = 0;
     for (int i = 0; i < token_num; i++)
       {
       if ( equal_(i, token) ) cnt ++;
       if (cnt == n) return i;
       }
     fprintf(stderr,"\nTokenString::token_index: less than %d instances of [%s]\n",n,token),
     exit(1);
     }

//***************************************************************************

double TokenString::trail_double(const char *token, int i)  
     { double d;
       trail_pars(token, i, 'd', &d, 'E');
       return d; }

double TokenString::trail2_double(const char *token1, const char *token2, int i)
     { double d;
       trail2_pars(token1, token2, i, 'd', &d, 'E');
       return d; }

//***************************************************************************

int TokenString::trail_int(const char *token, int i) 
     { int j;
       trail_pars(token, i, 'i', &j, 'E');
       return j; }

int TokenString::trail2_int(const char *token1, const char *token2, int i)
     { int j;
       trail2_pars(token1, token2, i, 'i', &j, 'E');
       return j; }

//***************************************************************************

long TokenString::trail_long(const char *token, int i) 
     { long j;
       trail_pars(token, i, 'l', &j, 'E');
       return j; }

long TokenString::trail2_long(const char *token1, const char *token2, int i)
     { long j;
       trail2_pars(token1, token2, i, 'l', &j, 'E');
       return j; }

//***************************************************************************

char TokenString::is_number(const char *s) 
   { 
   if (::is_number(s)) return 1;
     else if (token2_count(s,"=")) return 1;
       else return 0;
   }

//***************************************************************************

double TokenString::token_double(const char *s)
   {
   double r;
   int n;

   if (::is_number(s) ) sscanf(s,"%lf",&r);
   else if (n = token2_count(s,"="))
       {
       int p = token2_index(s,"=");
       r = KinExpression(*this, p+1, 0).Evaluate();
       if (n > 1) 
	 fprintf(stderr,"\n*** WARNING: TokenString::token_double: more than one definition for [%s] \n",s);
       }
   else fprintf(stderr,"\nTokenString::token_double: [%s] not a number\n",s), exit(1);

   return r;
   }

//***************************************************************************

int TokenString::token_int(const char *s)
   {
   int r;
   int  n;

   if (::is_number(s) ) sscanf(s,"%d",&r);
   else if (n = token2_count(s,"="))
       {
       int p = token2_index(s,"=");
       r = int( KinExpression(*this, p+1, 0).Evaluate() + 0.5 );
       if (n > 1) 
	 fprintf(stderr,"\n** TokenString::token_int: more than one definition for [%s] ****\n",s),
         exit(1);
       }
   else fprintf(stderr,"\nTokenString::token_int: [%s] not a number\n",s), exit(1);

   return r;
   }

//***************************************************************************

long TokenString::token_long(const char *s)
   {
   long r;
   int  n;

   if (::is_number(s) ) sscanf(s,"%ld",&r);
   else if (n = token2_count(s,"="))
       {
       int p = token2_index(s,"=");
       r = long( KinExpression(*this, p+1, 0).Evaluate() + 0.5 );
       if (n > 1) 
	 fprintf(stderr,"\n** TokenString::token_int: more than one definition for [%s] ****\n",s),
         exit(1);
       }
   else fprintf(stderr,"\nTokenString::token_long: [%s] not a number\n",s), exit(1);

   return r;
   }

//***************************************************************************

void TokenString::token_string(const char *sin, char *sout)
   {
   int  n = token2_count(sin,"=");
   strcpy(sout,"");

   if (n > 1) 
     {
     fprintf(stderr,"\n** TokenString::token_string: more than one definition for [%s] ****\n",sin);
     exit(1);
     }

   if ( n == 0 )  strcpy(sout,sin);  // simple string
   else
       {
       int p0 = token2_index(sin,"=");
       int m = tokens_to_eol(p0 + 1);

       for (int p = p0 + 1; p <= p0 + m; p++)
         if ( equal(p,"+") ) continue;
         else if (is_number(p)) 
           {
	   char s[40];
           double x = index_double(p);
           sprintf(s,"%.4g",x);
           strcat(sout,s);
	   }
         else strcat(sout, token_string( token_ptr[p] ));
       }
   }

//***************************************************************************

char *TokenString::token_string(const char *sin)
   {
     char sout[1024];
     token_string(sin, sout);
     return ::StrCpy(sout);
   }

//***************************************************************************

int TokenString::tokens_to_eol(int p)
{
  int    num = 0;

  while ( !equal(p, carr_ret) ) { p++; num++; }

  return num;
}

//***************************************************************************

